"errors"

import logging
from enum import Enum
from typing import Optional, Union

from volcenginesdkarkruntime._exceptions import (
    ArkAPITimeoutError,
    ArkRateLimitError,
)

from ark.core.idl.common_protocol import ArkError
from ark.core.utils import context


class ReprEnum(Enum):
    """
    Only changes the repr(), leaving str() and format() to the mixed-in type.
    """


class StrEnum(str, ReprEnum):
    """
    Enum where members are also (and must be) strings
    """

    def __new__(cls, *values: str) -> "StrEnum":
        "values must already be of type `str`"
        if len(values) > 3:
            raise TypeError("too many arguments for str(): %r" % (values,))
        if len(values) == 1:
            # it must be a string
            if not isinstance(values[0], str):
                raise TypeError("%r is not a string" % (values[0],))
        if len(values) >= 2:
            # check that encoding argument is a string
            if not isinstance(values[1], str):
                raise TypeError("encoding must be a string, not %r" % (values[1],))
        if len(values) == 3:
            # check that errors argument is a string
            if not isinstance(values[2], str):
                raise TypeError("errors must be a string, not %r" % (values[2]))
        value = str(*values)
        member = str.__new__(cls, value)
        member._value_ = value
        return member

    def _generate_next_value_(self, start, count, last_values) -> str:  # type: ignore
        """
        Return the lower-cased version of the member name.
        """
        return self.lower()


class ErrorCode(StrEnum):
    http_code: int
    message: str
    error_type: str

    def __new__(
        cls,
        code: str,
        http_code: int,
        message: Optional[str] = None,
        error_type: Optional[str] = None,
    ) -> "ErrorCode":
        error_code = str.__new__(cls, code)
        error_code._value_ = code
        error_code.http_code = http_code  # type: ignore
        error_code.message = message or ""  # type: ignore
        error_code.error_type = error_type or ""  # type: ignore
        return error_code

    MissingParameter = (
        "MissingParameter",
        400,
        "The request failed because it is missing required parameters",
        "BadRequest",
    )
    InvalidParameter = (
        "InvalidParameter",
        400,
        "A parameter specified in the request is not valid: {parameter}",
        "BadRequest",
    )
    InvalidAction = (
        "InvalidAction",
        400,
        "The request targeted action that does not exist or is invalid.",
        "BadRequest",
    )
    InvalidBot = (
        "InvalidBot",
        400,
        "The request targeted bot that does not exist or is invalid.",
        "BadRequest",
    )
    ResourceNotFound = (
        "ResourceNotFound",
        404,
        "The specified resource is not found",
        "NotFound",
    )
    RateLimitExceeded = (
        "RateLimitExceeded",
        429,
        "The Requests Per Minute(RPM) limit of the associated {resource_type} for your "
        "account has been exceeded.",
        "TooManyRequests",
    )
    ServerOverloaded = (
        "ServerOverloaded",
        429,
        "The service: {service} is currently unable to handle "
        "additional requests due to server overload.",
        "TooManyRequests",
    )

    SensitiveContentDetected = (
        "SensitiveContentDetected",
        400,
        "The request failed because the input text may contain sensitive information.",
        "BadRequest",
    )

    AuthenticationError = (
        "AuthenticationError",
        401,
        "The API key in the request is missing or invalid.",
        "Unauthorized",
    )

    AccessDenied = (
        "AccessDenied",
        403,
        "The request failed because you do not have access to the requested resource.",
        "Forbidden",
    )

    AccountOverdueError = (
        "AccountOverdueError",
        403,
        "The request failed because your account has an overdue balance.",
        "Forbidden",
    )
    QuotaExceeded = (
        "QuotaExceeded",
        429,
        "Your account {account_id} has exhausted its free trial "
        "quota for {resource_type}.",
        "TooManyRequests",
    )

    InternalServiceError = (
        "InternalServiceError",
        500,
        "The service encountered an unexpected internal error.",
        "InternalServerError",
    )

    Unknown = ("Unknown", 500, "Unknown error")

    ModelLoadingError = (
        "ModelLoadingError",
        429,
        "The request cannot be processed at this time because "
        "the model is currently being loaded",
        "TooManyRequests",
    )

    KnowledgeBaseError = (
        "KnowledgeBaseError",
        500,
        "Retrieving knowledgebase meet error",
        "InternalServerError",
    )

    LinkReaderBaseError = (
        "LinkReaderBaseError",
        500,
        "Reading links meet error",
        "InternalServerError",
    )

    APITimeoutError = (
        "APITimeoutError",
        500,
        "Request timed out",
        "InternalServerError",
    )


class APIException(Exception):
    def __init__(
        self,
        message: str,
        code: Union[ErrorCode, str],
        parameter: Optional[str] = None,
        http_code: Optional[int] = 500,
        error_type: Optional[str] = "InternalServerError",
    ):
        super().__init__(message)
        self.message = f"{message} Request id: {context.get_reqid()}"
        self.resource_id = context.get_resource_id()
        self.account_id = context.get_account_id() or ""
        self.code = code.value if isinstance(code, ErrorCode) else str(code)
        self.http_code = (
            code.http_code if isinstance(code, ErrorCode) else (http_code or 500)
        )
        self.type = (
            code.error_type
            if isinstance(code, ErrorCode)
            else (error_type or "InternalServerError")
        )
        self.parameter = parameter

    def __str__(self) -> str:
        return (
            "Detailed exception information is listed below.\n"
            + "account_id: {}\n"
            + "resource_id: {}\n"
            + "code: {}\n"
            + "message: {}"
        ).format(self.account_id, self.resource_id, self.code, self.message)

    def to_error(self) -> ArkError:
        return ArkError(
            code=self.code, message=self.message, param=self.parameter, type=self.type
        )


class InternalServiceError(APIException):
    def __init__(self, message: str):
        logging.error(f"[Internal Error]: {message}")
        super().__init__(
            message=ErrorCode.InternalServiceError.message,
            code=ErrorCode.InternalServiceError,
        )


class InvalidParameter(APIException):
    def __init__(self, parameter: str, cause: Optional[str] = None):
        message = ErrorCode.InvalidParameter.message.format(parameter=parameter)
        if cause:
            message = f"{message}.{cause}"
        super().__init__(
            message=message, code=ErrorCode.InvalidParameter, parameter=parameter
        )


class InvalidAction(APIException):
    def __init__(self, parameter: Optional[str] = None):
        message = ErrorCode.InvalidAction.message
        super().__init__(
            message=message, code=ErrorCode.InvalidAction, parameter=parameter
        )


class InvalidBot(APIException):
    def __init__(self, parameter: Optional[str] = None):
        message = ErrorCode.InvalidBot.message
        super().__init__(
            message=message, code=ErrorCode.InvalidBot, parameter=parameter
        )


class MissingParameter(APIException):
    def __init__(self, parameter: Optional[str] = None):
        message = (
            f"{ErrorCode.MissingParameter.message}:{parameter}"
            if parameter
            else str(ErrorCode.MissingParameter.message)
        )
        super().__init__(
            message=message, code=ErrorCode.MissingParameter, parameter=parameter
        )


class ResourceNotFound(APIException):
    def __init__(self, resource_type: Optional[str] = None):
        message = (
            f"{ErrorCode.ResourceNotFound.message}:{resource_type}"
            if resource_type
            else str(ErrorCode.ResourceNotFound.message)
        )
        super().__init__(
            message,
            code=ErrorCode.ResourceNotFound,
            parameter=resource_type,
        )


class RateLimitExceeded(APIException):
    def __init__(self, resource_type: str):
        message = ErrorCode.RateLimitExceeded.message.format(
            resource_type=resource_type
        )
        super().__init__(
            message,
            code=ErrorCode.RateLimitExceeded,
            parameter=resource_type,
        )


class ServerOverloaded(APIException):
    def __init__(self, service: str):
        message = ErrorCode.ServerOverloaded.message.format(service=service)
        super().__init__(
            message,
            code=ErrorCode.ServerOverloaded,
            parameter=service,
        )


class AuthenticationError(APIException):
    def __init__(self, cause: Optional[str] = None):
        message = ErrorCode.AuthenticationError.message
        if cause:
            message = f"{message} {cause}"
        super().__init__(
            message=message,
            code=ErrorCode.AuthenticationError,
        )


class AccessDenied(APIException):
    def __init__(self, cause: Optional[str] = None):
        message = ErrorCode.AccessDenied.message
        if cause:
            message = f"{message}.{cause}"
        super().__init__(
            message,
            code=ErrorCode.AccessDenied,
        )


class QuotaExceeded(APIException):
    def __init__(self, account_id: str, resource_type: str):
        message = ErrorCode.QuotaExceeded.message.format(
            account_id=account_id, resource_type=resource_type
        )
        super().__init__(message, code=ErrorCode.QuotaExceeded, parameter=resource_type)


class SensitiveContentDetected(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.SensitiveContentDetected)


class AccountOverdueError(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.AccountOverdueError)


class KnowledgeBaseError(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.KnowledgeBaseError)


class LinkReaderBaseError(APIException):
    def __init__(self, message: str):
        super().__init__(message, code=ErrorCode.LinkReaderBaseError)


class APITimeoutError(APIException):
    def __init__(self, message: str):
        logging.error(f"[API timeout error]: {message}")
        super().__init__(
            message=ErrorCode.APITimeoutError.message,
            code=ErrorCode.APITimeoutError,
        )


FALLBACK_EXCEPTIONS = (
    APITimeoutError,
    ArkAPITimeoutError,
    RateLimitExceeded,
    ArkRateLimitError,
)



from typing import Any, Dict, List

from langchain.load import dumps, loads
from langchain_core.documents import Document

from ark.core.task import task


@task(distributed=False)
def reciprocal_rank_fusion(
    results: List[Dict[str, Any]], k: float = 60
) -> List[Document]:
    fused_scores = {}
    for docs in results:
        # Assumes the docs are returned in sorted order of relevance
        for rank, doc in enumerate(docs):
            doc_str = dumps(doc)
            if doc_str not in fused_scores:
                fused_scores[doc_str] = 0.0
            fused_scores[doc_str] += 1.0 / (rank + k)

    reranked_results = [
        (loads(doc), score)
        for doc, score in sorted(fused_scores.items(), key=lambda x: x[1], reverse=True)
    ]
    context_docs = "\n".join([doc[0].page_content for doc in reranked_results])
    return [Document(page_content=context_docs)]



from typing import Any, Dict, List, Optional


def list_item_merge(
    a: List[Dict[str, Any]], b: List[Dict[str, Any]], unique_key: Optional[str]
) -> List[Dict[str, Any]]:
    """
    merge two list items into a new one
    will use unique_key to identify the item
    """
    if not unique_key:
        return a + b

    merged = []
    unique_dict = {bb.get(unique_key, ""): bb for bb in b}
    for item in a:
        if item.get(unique_key, "") not in unique_dict.keys():
            merged.append(item)

    merged += b

    return merged



import json
import os
import time
from typing import Any, Dict, List, Optional, Set, Union

from volcengine.ApiInfo import ApiInfo
from volcengine.base.Service import Service
from volcengine.Credentials import Credentials
from volcengine.maas.exception import MaasException, new_client_sdk_request_error
from volcengine.ServiceInfo import ServiceInfo

from ark.core.client.base import Client


class ArkAuthorization(Client, Service):
    def __init__(
        self,
        host: str = "open.volcengineapi.com",
        region: str = "cn-beijing",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        connection_timeout: int = 300,
        socket_timeout: int = 300,
        **kwargs: Any,
    ):
        ak = ak or os.getenv("VOLC_ACCESSKEY")
        sk = sk or os.getenv("VOLC_SECRETKEY")
        service_info: ServiceInfo = self.get_service_info(
            host, region, ak, sk, connection_timeout, socket_timeout
        )
        api_info: Dict[str, ApiInfo] = self.get_api_info()

        super().__init__(service_info=service_info, api_info=api_info, **kwargs)
        self.resource_ids: Set[str] = set()
        self._setted_apikey: Optional[str] = None
        self._apikey_ttl: int = 604800

        self.set_ak(ak)
        self.set_sk(sk)

    @staticmethod
    def get_service_info(
        host: str,
        region: str,
        ak: Optional[str],
        sk: Optional[str],
        connection_timeout: int,
        socket_timeout: int,
    ) -> ServiceInfo:
        service_info = ServiceInfo(
            host,
            {"Accept": "application/json"},
            Credentials(ak, sk, "ark", region),
            connection_timeout,
            socket_timeout,
            "https",
        )
        return service_info

    @staticmethod
    def get_api_info() -> Dict[str, ApiInfo]:
        api_info = {
            "get_apikey": ApiInfo(
                method="POST",
                path="/",
                query={"Action": "GetApiKey", "Version": "2024-01-01"},
                form={},
                header={},
            )
        }
        return api_info

    def get_apikey_ttl(self) -> int:
        return self._apikey_ttl

    def _set_apikey_ttl(self, ttl: int) -> None:
        self._apikey_ttl = ttl

    def generate_apikey(
        self,
        resource_ids: List[str],
        ttl: int = 604800,
        resource_type: str = "endpoint",
    ) -> str:
        assert len(resource_ids) > 0

        req = {
            "DurationSeconds": ttl,
            "ResourceType": resource_type,
            "ResourceIds": resource_ids,
        }
        try:
            res = self.json("get_apikey", {}, json.dumps(req))
            res = json.loads(res)
            apikey = res.get("Result", {}).get("ApiKey", "")
            self._setted_apikey = apikey

            self._set_apikey_ttl(ttl + int(time.time()))
            return apikey
        except MaasException as e:
            raise e
        except Exception as e:
            raise new_client_sdk_request_error(str(e))

    def get_invalid_resource_ids(
        self, resource_ids: Union[Set[str], List[str]]
    ) -> List[str]:
        resource_ids_set = set(resource_ids)
        invalid_resource_ids = resource_ids_set - self.resource_ids

        return list(invalid_resource_ids)

    def get_apikey(
        self,
        resource_ids: Union[List[str], Set[str]],
        ttl: int = 604800,
        resource_type: str = "endpoint",
    ) -> Optional[str]:
        # exists invalid resource id
        if (
            self._setted_apikey is None
            or len(self.get_invalid_resource_ids(resource_ids)) > 0
        ):
            self.generate_apikey(list(resource_ids), ttl, resource_type)
            self.resource_ids = set(resource_ids)

        # 过期5分钟前,重新生成
        if time.time() + 300 > self.get_apikey_ttl():
            self.generate_apikey(list(resource_ids), ttl, resource_type)
            self.resource_ids = set(resource_ids)

        return self._setted_apikey


class APIGAuthorization(Client, Service):
    def __init__(
        self,
        host: str = "open.volcengineapi.com",
        region: str = "cn-beijing",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        connection_timeout: int = 300,
        socket_timeout: int = 300,
        **kwargs: Any,
    ) -> None:
        ak = ak or os.getenv("VOLC_ACCESSKEY")
        sk = sk or os.getenv("VOLC_SECRETKEY")
        service_info = self.get_service_info(
            host, region, ak, sk, connection_timeout, socket_timeout
        )
        api_info = self.get_api_info()

        super().__init__(service_info=service_info, api_info=api_info, **kwargs)
        self._setted_apikey: Optional[str] = None
        self._apikey_ttl: int = 604800

        self.set_ak(ak)
        self.set_sk(sk)

    @staticmethod
    def get_service_info(
        host: str,
        region: str,
        ak: Optional[str],
        sk: Optional[str],
        connection_timeout: int,
        socket_timeout: int,
    ) -> ServiceInfo:
        service_info = ServiceInfo(
            host,
            {"Accept": "application/json"},
            Credentials(ak, sk, "apig", region),
            connection_timeout,
            socket_timeout,
            "https",
        )
        return service_info

    @staticmethod
    def get_api_info() -> Dict[str, ApiInfo]:
        api_info = {
            "get_jwttoken": ApiInfo(
                method="POST",
                path="/",
                query={"Action": "GetJwtToken", "Version": "2021-03-03"},
                form={},
                header={},
            )
        }
        return api_info

    def get_apikey_ttl(self) -> int:
        return self._apikey_ttl

    def _set_apikey_ttl(self, ttl: int) -> None:
        self._apikey_ttl = ttl

    def generate_apikey(
        self, service_id: Optional[str] = None, gateway_id: Optional[str] = None
    ) -> str:
        ttl: int = 604800
        req = {"GatewayId": gateway_id, "ServiceId": service_id}
        try:
            res = self.json("get_jwttoken", {}, json.dumps(req))
            res = json.loads(res)

            apikey = res.get("Result", {}).get("JwtToken", "")
            self._setted_apikey = apikey
            self._set_apikey_ttl(ttl + int(time.time()))

            return apikey
        except MaasException as e:
            raise e
        except Exception as e:
            raise new_client_sdk_request_error(str(e))

    def get_apikey(
        self, service_id: Optional[str] = None, gateway_id: Optional[str] = None
    ) -> str:
        assert (
            gateway_id or service_id
        ), "either gateway_id or service_id should be filled"

        if self._setted_apikey is None:
            return self.generate_apikey(service_id, gateway_id)

        if time.time() + 300 > self.get_apikey_ttl():  # 过期5分钟前,重新生成
            return self.generate_apikey(service_id, gateway_id)

        return self._setted_apikey



from typing import List

from langchain_core.output_parsers import BaseTransformOutputParser


class RagIntentMessageChunkOutputParser(BaseTransformOutputParser[bool]):
    """OutputParser that parses BaseMessageChunk into intent of whether to search."""

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this class is serializable."""
        return True

    @property
    def _type(self) -> str:
        """Return the output parser type for serialization."""
        return "default"

    def parse(self, text: str) -> bool:
        """Parse the output of the resource_id into a bool.

        Args:
            text (str): The output of the resource_id.

        Returns:
            Tuple(bool, str): The bool indicating whether to search.
                True, return empty str
                False, return direct answer
        """
        if "无需检索" in text:
            return False
        return True


class RagRewriteMessageChunkOutputParser(BaseTransformOutputParser[str]):
    """OutputParser that parses BaseMessageChunk into intent of whether to search."""

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this class is serializable."""
        return True

    @property
    def _type(self) -> str:
        """Return the output parser type for serialization."""
        return "default"

    def parse(self, text: str) -> str:
        """Parse the output of the resource_id into a bool."""
        return text.strip()


class RagRewriteOutputParser(BaseTransformOutputParser[List[str]]):
    """Parse llm output to List[str]"""

    def parse(self, text: str) -> List[str]:
        queries = text.split("\n")
        # filter extreme shot query
        return [q.strip() for q in queries if len(q) > 3]



import json
import logging
from typing import Any, Dict, List, Tuple

from ark.core.client import ArkClient, get_client_pool
from ark.core.idl.common_protocol import Action, Function, Server
from ark.core.idl.maas_protocol import Tool
from ark.core.utils.errors import InvalidParameter

ToolTypeFunction = "function"


def get_example_values(examples: Dict[str, Any]) -> Dict[str, Any]:
    values_list = list(examples.values())
    json_examples = {}
    for value_dict in values_list:
        if "value" in value_dict:
            json_examples.update(value_dict["value"])
    return json_examples


def merge_parameters(
    params_schema: Dict[str, Any],
    examples: Dict[str, Any],
    parameters_list: List[Dict[str, Any]],
) -> None:
    for parameter_info in parameters_list:
        name = parameter_info.get("name", "")
        schema_info = parameter_info.get("schema", {})
        property_schema = {
            "type": schema_info.get("type", ""),
            "description": parameter_info.get("description", ""),
        }
        params_schema.get("properties", {})[name] = property_schema

        if parameter_info.get("required", False):
            required_list = params_schema.get("required", [])
            if name not in required_list:
                required_list.append(name)
            params_schema["required"] = required_list

        if "example" in schema_info:
            examples[name] = get_example_values(schema_info.get("example", {}))


def resolve_ref(ref: str, components: Dict[str, Any]) -> Dict[str, Any]:
    if not ref.startswith("#/components/schemas/"):
        logging.error(f"Unsupported reference: {ref}")
        raise InvalidParameter(f"Unsupported reference: {ref}")

    schema_name = ref[len("#/components/schemas/") :]
    if schema_name in components:
        return components[schema_name]
    else:
        logging.error(f"Schema not found in components: {schema_name}")
        raise InvalidParameter(f"Schema not found in components: {schema_name}")


def resolve_refs_in_request_body(
    request_body: Dict[str, Any], components: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    content = request_body.get("content", {})
    for media_type, media_info in content.items():
        schema_ref = media_info.get("schema", {}).get("$ref", None)
        if not schema_ref:
            continue

        resolved_schema = resolve_ref(schema_ref, components)
        if not resolved_schema:
            continue

        properties = resolved_schema.get("properties", {})
        for name, property in properties.items():
            if isinstance(property, dict):
                properties[name].pop("required", None)

            media_info["schema"] = resolved_schema

    json_content = content.get("application/json")
    return json_content.get("schema", {}), get_example_values(
        json_content.get("examples", {})
    )


def convert_actions(
    actions: List[Dict[str, Any]],
) -> Dict[str, Action]:
    actions_map: Dict[str, Action] = {}
    for raw_action in actions:
        servers = raw_action.get("servers", [])
        paths = raw_action.get("paths", [])

        try:
            assert len(servers) > 0, "servers are required"
            assert len(paths) > 0, "paths are required"
        except AssertionError as e:
            raise InvalidParameter(f"Invalid browsing_options: {e}")

        components = raw_action.get("components", {}).get("schemas", {})
        for path, methods in paths.items():
            for method, details in methods.items():
                request_body = details.get("requestBody", {})
                ref_parameters, ref_examples = resolve_refs_in_request_body(
                    request_body, components
                )
                merge_parameters(
                    ref_parameters, ref_examples, details.get("parameters", [])
                )
                function = Function(
                    name=details.get("operationId", ""),
                    description=details.get("summary", ""),
                    parameters=ref_parameters,
                    examples=[json.dumps(ref_examples)],
                )

                for server_info in servers:
                    action = Action(
                        server=Server(
                            url=server_info.get("url", None),
                            description=server_info.get("description", None),
                        ),
                        http_method=method,
                        path=path,
                        tool=Tool(
                            type=ToolTypeFunction,
                            function=function,
                        ),
                    )
                    actions_map[function.name] = action

    return actions_map


def get_ark_client() -> ArkClient:
    client_pool = get_client_pool()
    client: ArkClient = client_pool.get_client("chat")  # type: ignore
    if not client:
        client = ArkClient()
    return client



import os
from typing import Any, Optional

from volcengine.viking_db import VikingDBService

from ark.core.client import Client


class VikingDBClient(Client, VikingDBService):
    def __init__(
        self,
        host: str = "api-vikingdb.volces.com",
        region: str = "cn-north-1",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            host=host,
            region=region,
            **kwargs,
        )
        self.set_ak(ak or os.getenv("VOLC_ACCESSKEY"))
        self.set_sk(sk or os.getenv("VOLC_SECRETKEY"))



from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

from pydantic import BaseModel, Field

from ark.component.action.utils import ToolTypeFunction, get_ark_client
from ark.core.client import ArkClient
from ark.core.idl.common_protocol import Function, Parameters, Tool
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
    MaasChatRequest,
    MaasChatResponse,
)
from ark.core.task import task


class FunctionCall(BaseModel):
    endpoint_id: str
    messages: List[MaasChatMessage]
    functions: List[Function]
    parameters: Optional[Union[Parameters, Dict[str, Any]]] = Field(
        default_factory=dict
    )
    client: ArkClient = Field(default_factory=get_ark_client)
    extra: Optional[Dict[str, str]] = Field(default_factory=dict)
    _tools: Optional[List[Tool]] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @property
    def tools(self) -> List[Tool]:
        if not self._tools:
            self._tools = [
                Tool(type=ToolTypeFunction, function=function)
                for function in self.functions
            ]
        return self._tools

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasChatRequest,
    ) -> Union[AsyncIterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return await self.client.async_stream_chat(
                endpoint_id=endpoint_id, req=request
            )
        else:
            return await self.client.async_chat(endpoint_id=endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasChatRequest
    ) -> Union[Iterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return self.client.stream_chat(endpoint_id=endpoint_id, req=request)
        else:
            return self.client.chat(endpoint_id=endpoint_id, req=request.model_dump())

    async def arun(self) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.messages,
            tools=self.tools,
            parameters=self.parameters,
            extra=self.extra,
        )

        return await self._arun(self.endpoint_id, request)

    async def astream(self) -> AsyncIterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.messages,
            tools=self.tools,
            parameters=self.parameters,
            extra=self.extra,
        )

        async for resp in await self._arun(self.endpoint_id, request):
            yield resp

    def run(self) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.messages,
            tools=self.tools,
            parameters=self.parameters,
            extra=self.extra,
        )
        return self._run(self.endpoint_id, request)

    def stream(self) -> Iterator[dict]:
        request = MaasChatRequest(
            stream=True,
            messages=self.messages,
            tools=self.tools,
            parameters=self.parameters,
            extra=self.extra,
        )

        for resp in self._run(self.endpoint_id, request):
            yield resp



import logging
from urllib.parse import urlparse

import tos
from volcengine.Credentials import Credentials

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


def download_tos_file(
    url: str,
    credentials: Credentials,
    cache_path: str = "/tmp/",
    region: str = "cn-beijing",
) -> str:
    result = urlparse(url)
    if result.scheme == "tos":
        endpoint = "tos-cn-beijing.volces.com"
        bucket_name = result.netloc
        object_key = result.path[1:]
        file_name = object_key.split("/")[-1]
        local_path = cache_path + file_name
    elif result.scheme == "http":
        endpoint = ".".join(result.netloc.split(".")[1:])
        bucket_name = result.netloc.split(".")[0]
        object_key = result.path[1:]
        file_name = object_key.split("/")[-1]
        local_path = cache_path + file_name
    else:
        raise ValueError("url is not tos or http")
    try:
        # 创建 TosClientV2 对象，对桶和对象的操作都通过 TosClientV2 实现
        client = tos.TosClientV2(credentials.ak, credentials.sk, endpoint, region)
        # 若 file_name 为目录则将对象下载到此目录下, 文件名为对象名
        client.get_object_to_file(bucket_name, object_key, local_path)
    except tos.exceptions.TosClientError as e:
        # 操作失败，捕获客户端异常，一般情况为非法请求参数或网络异常
        LOGGER.error(
            "fail with client error, message:%s, cause: %s", e.message, e.cause
        )
    except tos.exceptions.TosServerError as e:
        # 操作失败，捕获服务端异常，可从返回信息中获取详细错误信息
        LOGGER.error("fail with server error, code: %s", e.code)
        # request id 可定位具体问题，强烈建议日志中保存
        LOGGER.error("error with request id: %s", e.request_id)
        LOGGER.error("error with message: %s", e.message)
        LOGGER.error("error with http code: %s", e.status_code)
        LOGGER.error("error with ec: %s", e.ec)
        LOGGER.error("error with request url: %s", e.request_url)
    except Exception as e:
        LOGGER.error("fail with unknown error: %s", e)
    return local_path


def pre_sign_tos_url(
    url: str, credentials: Credentials, region: str = "cn-beijing"
) -> str:
    # tos开始的做个签名,http 开始的默认已经做过签名
    result = urlparse(url)
    if result.scheme == "tos":
        try:
            # 创建 TosClientV2 对象，对桶和对象的操作都通过 TosClientV2 实现
            client = tos.TosClientV2(
                credentials.ak,
                credentials.sk,
                security_token=credentials.session_token,
                endpoint="tos-cn-beijing.volces.com",
                region=region,
            )
            # 生成上传文件的签名url，有效时间为3600s
            out = client.pre_signed_url(
                tos.HttpMethodType.Http_Method_Get,
                bucket=result.netloc,
                key=result.path[1:],
                expires=3600,
            )
            signed_url = out.signed_url
            # LOGGER.info("signed_url: %s", signed_url)
            return signed_url

        except tos.exceptions.TosClientError as e:
            # 操作失败，捕获客户端异常，一般情况为非法请求参数或网络异常
            LOGGER.error(
                "fail with client error, message:%s, cause: %s", e.message, e.cause
            )
        except tos.exceptions.TosServerError as e:
            # 操作失败，捕获服务端异常，可从返回信息中获取详细错误信息
            LOGGER.error("fail with server error, code: %s", e.code)
            # request id 可定位具体问题，强烈建议日志中保存
            LOGGER.error("error with request id: %s", e.request_id)
            LOGGER.error("error with message: %s", e.message)
            LOGGER.error("error with http code: %s", e.status_code)
            LOGGER.error("error with ec: %s", e.ec)
            LOGGER.error("error with request url: %s", e.request_url)
        except Exception as e:
            LOGGER.error("fail with unknown error: %s", e)

    return ""



from typing import Any, AsyncIterable, Dict, List, Tuple, Type, Union

from pydantic import Field, field_validator

from ark.component.v3.search.config import SearchConfig
from ark.core.action import Action
from ark.core.idl.ark_protocol import (
    ArkChatCompletionChunk,
    ArkChatRequest,
    ArkChatResponse,
    ArkSearchIntentionMetadata,
    ArkSearchRequest,
    ArkSearchResponse,
    ArkSearchSourceType,
    ArkSearchSummaryMetadata,
    SourceType,
)
from ark.core.idl.common_protocol import ActionDetails
from ark.core.task.task import task
from ark.core.utils.errorsv3 import InvalidParameter


class Search(Action[ArkSearchRequest, ArkSearchResponse]):
    name: str = "Search"
    response_cls: Type[ArkSearchResponse] = ArkSearchResponse

    @classmethod
    @field_validator("meta_info")
    def validate_meta_info(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        if v.get("action_name") not in [
            "WebBrowsingPlus",
            "WebBrowsingPro",
            "content_plugin",
        ]:
            raise InvalidParameter("action_name")
        return v

    @task()
    async def arun(
        self, request: ArkSearchRequest, **kwargs: Any
    ) -> Union[ArkSearchResponse, Tuple[ArkSearchResponse, ActionDetails]]:
        if self.meta_info.action_name:
            request.action_name = self.meta_info.action_name  # type: ignore

        return await super().arun(request, **kwargs)


class SearchIntention(Action[ArkChatRequest, ArkChatResponse]):
    name: str = "SearchIntention"
    meta_info: ArkSearchIntentionMetadata = Field(
        default_factory=ArkSearchIntentionMetadata
    )
    response_cls: Type[ArkChatResponse] = ArkChatResponse

    @classmethod
    @field_validator("meta_info")
    def validate_meta_info(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        if v.get("action_name") not in [
            "WebBrowsing",
            "WebBrowsingPlus",
            "WebBrowsingPro",
            "content_plugin",
        ]:
            raise InvalidParameter("action_name")
        return v

    @task()
    async def arun(self, request: ArkChatRequest, **kwargs: Any) -> ArkChatResponse:
        request.metadata = self.meta_info.model_dump(exclude_none=True)
        if not self.meta_info.model:
            self.meta_info.model = request.model
        return await super().arun(request, **kwargs)

    @staticmethod
    def _default_result_mappings() -> Dict[str, bool]:
        return {"需要": True, "不需要": False}

    @task()
    async def aparse_output(self, text: str) -> bool:
        _result_mappings = (
            self.meta_info.result_mapping or self._default_result_mappings()
        )
        # default false
        return _result_mappings.get(text, False)


class SearchSummary(Action[ArkChatRequest, ArkChatResponse]):
    name: str = "SearchSummary"
    meta_info: ArkSearchSummaryMetadata = Field(
        default_factory=ArkSearchIntentionMetadata
    )
    response_cls: Type[ArkChatResponse] = ArkChatResponse

    @classmethod
    @field_validator("meta_info")
    def validate_meta_info(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        if v.get("action_name") not in [
            "WebBrowsing",
            "WebBrowsingPlus",
            "WebBrowsingPro",
            "content_plugin",
        ]:
            raise InvalidParameter("action_name")
        return v

    @task()
    async def arun(self, request: ArkChatRequest, **kwargs: Any) -> ArkChatResponse:
        request.metadata = self.meta_info.model_dump(exclude_none=True)

        if not self.meta_info.model:
            self.meta_info.model = request.model

        return await super().arun(request, **kwargs)

    @task()
    async def astream(
        self, request: ArkChatRequest, **kwargs: Any
    ) -> AsyncIterable[ArkChatCompletionChunk]:
        request.metadata = self.meta_info.model_dump(exclude_none=True)
        self.meta_info.model = request.model

        responses = self.client.astream_request(
            api=self.name,
            meta_info=self.meta_info,
            request=request,
            response_cls=ArkChatCompletionChunk,
            **kwargs,
        )
        async for response in responses:
            yield response


@task()
def get_extended_source_types(search_config: SearchConfig) -> List[SourceType]:
    source_type_list: List[SourceType] = []
    if not search_config.source_type:
        return source_type_list

    for source_type in search_config.source_type:
        if source_type == ArkSearchSourceType.DEFAULT_SEARCH:
            continue
        source_type_list.append(source_type)
    return source_type_list



import os
from typing import Any, Dict, Optional, Tuple, Type

from ark.core.bot.server import BotServer
from ark.core.client import Client, get_default_client_configs
from ark.core.launcher.runner import get_endpoint_config, get_runner
from ark.core.utils.context import set_account_id, set_resource_id, set_resource_type
from ark.core.utils.package_management import load_function
from ark.telemetry.trace.setup import TraceConfig, setup_tracing


def launch_serve(
    package_path: str,
    host: str = "0.0.0.0",
    port: int = 8080,
    endpoint_path: str = "/api/v3/bots/chat/completions",
    health_check_path: Optional[str] = "/healthz",
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
    trace_config: Optional[TraceConfig] = None,
    trace_on: bool = True,
    **kwargs: Any,
) -> None:
    set_resource_type(os.getenv("RESOURCE_TYPE") or "")
    set_resource_id(os.getenv("RESOURCE_ID") or "")
    set_account_id(os.getenv("ACCOUNT_ID") or "")

    setup_tracing(
        endpoint=os.getenv("TRACE_ENDPOINT"),
        trace_config=trace_config,
        trace_on=trace_on,
    )

    runnable_func = load_function(package_path, "main")

    server: BotServer = BotServer(
        runner=get_runner(runnable_func),
        health_check_path=health_check_path,
        endpoint_config=get_endpoint_config(endpoint_path, runnable_func),
        clients=clients if clients else get_default_client_configs(),
    )
    server.run(app=server.app, host=host, port=port, **kwargs)



from ark.component.prompts.browsing_prompt import (
    BrowsingGenerationChatPromptTemplate,
    BrowsingIntentionChatPromptTemplate,
)
from ark.component.prompts.custom_prompt import CustomPromptTemplate

__all__ = [
    "BrowsingGenerationChatPromptTemplate",
    "BrowsingIntentionChatPromptTemplate",
    "CustomPromptTemplate",
]



from typing import Any, List

from pydantic import Field
from ark.component.v3.llm.base import BaseChatPromptTemplate
from ark.core.idl.ark_protocol import ArkGroupChatConfig
from langchain_core.messages import (BaseMessage, AIMessage, HumanMessage, SystemMessage)


class GroupChatPromptTemplate(BaseChatPromptTemplate):
    group_chat_config: ArkGroupChatConfig = Field(default_factory=ArkGroupChatConfig)
    """ History QA Messages """
    input_variables: List[str] = ["messages", "target_character_name", ""]

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.

        Args:
            messages: list of original messages.
            target_character_name: chat target character name.

        Returns:
            Formatted message
        """
        if "messages" not in kwargs:
            raise ValueError("messages")
        if "target_character_name" not in kwargs:
            raise ValueError("target_character_name")

        messages = kwargs.pop("messages")
        input_messages: List[BaseMessage] = [
            msg.copy(deep=True) for msg in messages if msg.type in {"human", "ai"}
        ]

        # cast message with character name
        target_character_name = kwargs.pop("target_character_name")

        formatted_messages: List[BaseMessage] = []
        for msg in input_messages:
            if isinstance(msg, HumanMessage):
                formatted_messages.append(
                    HumanMessage(content=f"{self.group_chat_config.user_name or '用户'}：{msg.content}"))
            elif isinstance(msg, AIMessage):
                if msg.name == target_character_name:
                    """ character self """
                    formatted_messages.append(AIMessage(content=msg.content))
                else:
                    """ other character would be translated into human message"""
                    formatted_messages.append(HumanMessage(content=f"{msg.name}：{msg.content}"))

        target_character_config = None
        for character in self.group_chat_config.characters:
            if character.name == target_character_name:
                target_character_config = character
                break

        if not target_character_config:
            raise ValueError(f"target_character_name: {target_character_name} not found in group_chat_config")

        character_names = [c.name for c in self.group_chat_config.characters if c.name != target_character_name]
        if self.group_chat_config.user_name:
            character_names.append(self.group_chat_config.user_name)

        # inject system prompt
        sp = f"""你是{target_character_name}，{target_character_config.system_prompt}
        >>>>>>>>>>
        你现在正处于一个群聊环境，"""
        if self.group_chat_config.description:
            sp += f"群聊的背景信息如下：\n{self.group_chat_config.description}\n"
        sp += f"""以下是参与群聊的人员：
        【{character_names}】"""

        formatted_messages.insert(0, SystemMessage(content=sp))

        if isinstance(formatted_messages[-1], AIMessage):
            formatted_messages.append(AIMessage(content=""))

        return formatted_messages



import logging
import os
from typing import Optional

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource, ResourceAttributes
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SpanExporter,
)
from pydantic import BaseModel


class TraceConfig(BaseModel):
    # trace basic config
    ak: Optional[str] = None
    sk: Optional[str] = None
    topic: Optional[str] = None
    region: Optional[str] = None

    # batch exporter config
    max_queue_size: Optional[int] = None
    schedule_delay_millis: Optional[float] = None
    max_export_batch_size: Optional[int] = None
    export_timeout_millis: Optional[float] = None

    def __init__(
        self,
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        topic: Optional[str] = None,
        region: Optional[str] = None,
        max_queue_size: Optional[int] = None,
        schedule_delay_millis: Optional[float] = None,
        max_export_batch_size: Optional[int] = None,
        export_timeout_millis: Optional[float] = None,
    ):
        super().__init__(
            ak=ak or os.getenv("VOLC_ACCESSKEY", ""),
            sk=sk or os.getenv("VOLC_SECRETKEY", ""),
            topic=topic or os.getenv("TRACE_TOPIC", ""),
            region=region or os.getenv("REGION", "cn-beijing"),
            max_queue_size=max_queue_size,
            schedule_delay_millis=schedule_delay_millis,
            max_export_batch_size=max_export_batch_size,
            export_timeout_millis=export_timeout_millis,
        )


def setup_tracing(
    endpoint: Optional[str] = None,
    trace_on: bool = True,
    trace_config: Optional[TraceConfig] = None,
) -> None:
    if not trace_on:
        return

    # ensure only initialize once
    provider = trace._TRACER_PROVIDER
    if provider is not None:
        return

    exporter: SpanExporter = ConsoleSpanExporter()
    resource: Resource = Resource.create(
        {
            ResourceAttributes.SERVICE_NAME: "bot",
            ResourceAttributes.HOST_NAME: _get_host_name(),
        },
        schema_url="https://opentelemetry.io/schemas/1.4.0",
    )

    if not trace_config:
        trace_config = TraceConfig()

    if endpoint:
        headers = {
            "x-tls-otel-tracetopic": trace_config.topic or os.getenv("TRACE_TOPIC", ""),
            "x-tls-otel-ak": trace_config.ak or os.getenv("VOLC_ACCESSKEY", ""),
            "x-tls-otel-sk": trace_config.sk or os.getenv("VOLC_SECRETKEY", ""),
            "x-tls-otel-region": trace_config.region
            or os.getenv("REGION", "cn-beijing"),
        }
        logging.info(f"initialize tls trace info: {headers}")
        exporter = OTLPSpanExporter(endpoint=endpoint, insecure=True, headers=headers)  # type: ignore

    provider = TracerProvider(resource=resource)
    provider.add_span_processor(
        BatchSpanProcessor(
            exporter,
            max_queue_size=trace_config.max_queue_size,  # type: ignore
            schedule_delay_millis=trace_config.schedule_delay_millis,  # type: ignore
            max_export_batch_size=trace_config.max_export_batch_size,  # type: ignore
            export_timeout_millis=trace_config.export_timeout_millis,  # type: ignore
        )
    )
    trace.set_tracer_provider(provider)


def _get_host_name() -> str:
    # default env key
    host_name = os.getenv("HOSTNAME", "")
    if not host_name and os.getenv("IS_LOCAL") is None:
        # faas env key
        host_name = os.getenv("_BYTEFAAS_POD_NAME", "")
    return host_name



from typing import Any, List

from langchain_core.output_parsers import BaseTransformOutputParser
from pydantic.v1 import Field
from typing_extensions import Literal

from ark.component.output_parser.rag_output import RagIntentMessageChunkOutputParser
from ark.component.prompts.rag_prompt import (
    ABSTRACT_INTENTION_TEMPLATE,
    KNOWLEDGE_INTENTION_TEMPLATE,
    IntentionPromptTemplate,
)
from ark.component.v3.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.core.task import task


def _knowledge_intention_prompt() -> IntentionPromptTemplate:
    return IntentionPromptTemplate(template=KNOWLEDGE_INTENTION_TEMPLATE)


def _abstract_intention_prompt() -> IntentionPromptTemplate:
    return IntentionPromptTemplate(template=ABSTRACT_INTENTION_TEMPLATE)


def _default_intention_parser() -> RagIntentMessageChunkOutputParser:
    return RagIntentMessageChunkOutputParser()


class RagIntention(BaseChatLanguageModel):
    name: Literal[
        "RagIntention", "KnowledgeIntention", "AbstractIntention"
    ] = "RagIntention"
    template: BaseChatPromptTemplate
    output_parser: BaseTransformOutputParser[List[bool]] = Field(
        default_factory=_default_intention_parser
    )

    def _get_prompt_template(self) -> BaseChatPromptTemplate:
        return self.template

    def _get_output_parser(self) -> BaseTransformOutputParser:
        return self.output_parser

    @task()
    async def arun(self, *args: Any, **kwargs: Any) -> bool:
        resp = await super().arun(*args, **kwargs)
        return await self.aparse_output(resp.choices[0].message.content)


class KnowledgeIntention(RagIntention):
    name: Literal["KnowledgeIntention"] = "KnowledgeIntention"
    template: BaseChatPromptTemplate = Field(
        default_factory=_knowledge_intention_prompt
    )


class AbstractIntention(RagIntention):
    name: Literal["AbstractIntention"] = "AbstractIntention"
    template: BaseChatPromptTemplate = Field(default_factory=_abstract_intention_prompt)



from typing import Any, Dict, List, Optional, Union

from pydantic import Field, field_validator

from ark.component.plugins.plugin import Plugin, convert_messages_to_keywords
from ark.core._api.deprecation import deprecated
from ark.core.client import ArkClient
from ark.core.idl.common_protocol import Parameters, Tool
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
)


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.search.SearchSummary",
)
class SearchSummary(Plugin):
    _tools: Optional[List[Tool]] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        endpoint_id: str,
        messages: List[MaasChatMessage] = Field(default_factory=list),
        options: Optional[Dict[str, Any]] = None,
        parameters: Optional[Union[Parameters, Dict[str, Any]]] = None,
        client: Optional[ArkClient] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name="SearchSummary",
            endpoint_id=endpoint_id,
            messages=messages,
            options=options,
            parameters=parameters,
            **kwargs,
        )
        if client:
            self.client = client

        if options:
            keywords = options.get("keywords", [])

            if len(keywords) == 0:
                options["keywords"] = [convert_messages_to_keywords(messages)[-1]]

        self._tools = [Tool(type=self.name, options=options)]

    @field_validator("options")
    @classmethod
    def validate_options(cls, v: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
        if not v or "keywords" not in v:
            return v

        keywords = v.get("keywords", [])

        assert isinstance(keywords, list) and all(
            isinstance(item, str) for item in keywords
        ), "keywords should be list of str"

        return v



import asyncio
import time
from typing import Any, Generic, Optional, Tuple, Type, Union

from httpx import Timeout
from pydantic import BaseModel, Field
from volcenginesdkarkruntime._exceptions import (
    ArkAPIError,
)

from ark.core.client import ArkActionClient, Client, get_client_pool
from ark.core.idl.ark_protocol import ArkActionMeta
from ark.core.idl.common_protocol import (
    ActionDetails,
    ExceptionInfo,
    RequestType,
    ResponseType,
    ToolDetails,
    ToolOutput,
    ToolOutputType,
)
from ark.core.task.task import task
from ark.core.utils.errorsv3 import (
    FALLBACK_EXCEPTIONS,
    APIException,
    APITimeoutError,
    InternalServiceError,
)


def _get_ark_client() -> Optional[Client]:
    client_pool = get_client_pool()
    client = client_pool.get_client("action")
    if not client:
        client = ArkActionClient(timeout=Timeout(connect=1.0, timeout=60.0))
    return client


def _ns_to_ms(ns: int) -> int:
    return int(ns / 1e6)


class Action(BaseModel, Generic[RequestType, ResponseType]):
    client: ArkActionClient = Field(default_factory=_get_ark_client)
    name: str
    meta_info: ArkActionMeta = Field(default_factory=ArkActionMeta)
    response_cls: Type[ResponseType]

    class Config:
        arbitrary_types_allowed = True

    def run(self, request: RequestType, **kwargs: Any) -> ResponseType:
        raise NotImplementedError

    @task(watch_io=False)
    async def arun(
        self, request: RequestType, **kwargs: Any
    ) -> Union[ResponseType, Tuple[ResponseType, ActionDetails]]:
        created_at = _ns_to_ms(time.time_ns())
        exception: Optional[Exception] = None
        try:
            response = await self.client.arequest(
                api=self.name,
                meta_info=self.meta_info,
                request=request,
                response_cls=self.response_cls,
                extra_headers=kwargs.get("extra_headers"),
                extra_query=kwargs.get("extra_query"),
                extra_body=kwargs.get("extra_body"),
                timeout=kwargs.get("timeout"),
            )
        except Exception as e:
            if isinstance(e, asyncio.TimeoutError):
                exception = APITimeoutError(str(e))
            else:
                exception = e
        finally:
            completed_at = _ns_to_ms(time.time_ns())

        if "include_details" in kwargs and kwargs["include_details"] is True:
            action_details = ActionDetails(
                name=self.meta_info.action_name
                if self.meta_info.action_name
                else self.name,
                count=1,
                tool_details=[
                    ToolDetails(
                        name=self.name,
                        input=request.model_dump(exclude_unset=True, exclude_none=True),
                        output=response.model_dump(
                            exclude_unset=True, exclude_none=True
                        )
                        if exception is None
                        else ToolOutput(
                            type=ToolOutputType.EXCEPTION,
                            data=ExceptionInfo(
                                type=exception.__class__.__name__,
                                message=exception.message
                                if isinstance(exception, (ArkAPIError, APIException))
                                else InternalServiceError(str(exception)).message,
                            ),
                        ),
                        created_at=created_at,
                        completed_at=completed_at,
                    )
                ],
            )
            if exception is not None and hasattr(exception, "message"):
                exception.message = action_details.model_dump_json(exclude_none=True)

        if exception is not None:
            if kwargs.get("special_exception_fallback") and (
                isinstance(exception, FALLBACK_EXCEPTIONS)
            ):
                response = self.response_cls()
            else:
                raise exception

        if "include_details" in kwargs and kwargs["include_details"] is True:
            return response, action_details
        return response



import json
import logging
import warnings
from typing import Any, Dict, List, Set, Union

from langchain.prompts.chat import BaseChatPromptTemplate
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
)
from langchain_core.messages.tool import ToolCall
from typing_extensions import Literal

from ark.core.idl.ark_protocol import ArkMessage, ChatCompletionMessageToolCallParam
from ark.core.idl.common_protocol import ChatRole
from ark.core.idl.maas_protocol import MaasChatMessage
from ark.core.task import task
from ark.core.utils.errorsv3 import InvalidParameter

MessageRole2ChatRole = {
    "human": ChatRole.USER,
    "ai": ChatRole.ASSISTANT,
    "system": ChatRole.SYSTEM,
    "function": ChatRole.FUNCTION,
}


def _convert_message_role_to_ark_role(  # type: ignore
    message_role: str,
) -> Literal["user", "system", "assistant", "tool"]:  # type: ignore
    if message_role == "human":
        return "user"
    elif message_role == "ai":
        return "assistant"
    elif message_role == "system":
        return "system"
    elif message_role == "tool" or message_role == "function":
        return "tool"


def maas_chat_messages_to_messages(
    chat_messages: Union[List[MaasChatMessage], List[ArkMessage]],
) -> List[BaseMessage]:
    """Convert llmflow engine chat request to langchain list of messages"""
    messages: List[BaseMessage] = []

    # Check if messages is None
    if chat_messages is None:
        raise ValueError("messages")

    for msg in chat_messages:
        if msg.content is None:
            warnings.warn(
                "ChatMessage contains content as None, replace it with empty str"
            )
            msg.content = ""

        kwargs = {"choice": {"message": msg.dict()}}
        if not isinstance(msg.content, str):
            continue
        if msg.role == ChatRole.ASSISTANT:
            messages.append(AIMessage(content=msg.content, additional_kwargs=kwargs))
        elif msg.role == ChatRole.SYSTEM:
            messages.append(
                SystemMessage(content=msg.content, additional_kwargs=kwargs)
            )
        elif msg.role == ChatRole.USER:
            messages.append(HumanMessage(content=msg.content, additional_kwargs=kwargs))
        elif msg.role == ChatRole.FUNCTION or msg.role == ChatRole.TOOL:
            messages.append(
                FunctionMessage(
                    name=msg.name if "name" in msg.model_fields and msg.name else "",
                    content=msg.content,
                    additional_kwargs=kwargs,
                )
            )
        else:
            raise ValueError(f"role: {msg.role}")
    return messages


def _convert_ark_messages(chat_messages: List[ArkMessage]) -> List[BaseMessage]:
    messages: List[BaseMessage] = []
    tool_calls: Dict[str, ChatCompletionMessageToolCallParam] = {}
    next_tool_ids: Set[str] = set()
    for message in chat_messages:
        if len(next_tool_ids) > 0 and message.role != "tool":
            raise InvalidParameter(
                parameter="messages",
                cause="An assistant message with 'tool_calls' must "
                "be followed by tool messages responding to each 'tool_call_id'.",
            )

        if message.role == "user":
            if isinstance(message.content, str):
                messages.append(
                    HumanMessage(content=message.content, name=message.name)
                )
            else:
                messages.append(
                    HumanMessage(
                        content=[
                            part.model_dump(exclude_none=True, exclude_unset=True)
                            for part in message.content
                        ]
                    )
                )

        elif message.role == "assistant":
            next_tool_ids = set()

            def load_arguments(arguments: str) -> Dict[str, Any]:
                try:
                    args = json.loads(arguments)
                    if isinstance(args, dict) and all(
                        isinstance(key, str) for key in args.keys()
                    ):
                        return args
                    return {}
                except json.JSONDecodeError:
                    logging.error(
                        f"json decode arguments failed: arguments={arguments}"
                    )
                    return {}

            content, thought = (message.content or ""), ""
            if message.tool_calls is not None:
                if isinstance(message.content, str):
                    sep = message.content.rfind("\n")
                    if sep >= 0:
                        content, thought = (
                            message.content[:sep],
                            message.content[sep + 1 :],
                        )

                for tool_call in message.tool_calls:
                    tool_calls[tool_call.id] = tool_call
                    next_tool_ids.add(tool_call.id)

            messages.append(
                AIMessage(
                    content=content,  # type: ignore
                    name=message.name,
                    tool_calls=[
                        ToolCall(
                            name=tool_call.function.name,
                            args=(
                                load_arguments(tool_call.function.arguments)
                                if tool_call.function.arguments
                                else {}
                            ),
                            id=tool_call.id,
                        )
                        for tool_call in (message.tool_calls or [])
                    ],
                    additional_kwargs={
                        "choice": {
                            "message": {
                                "tool_calls": [
                                    {
                                        "id": tool_call.id,
                                        "type": tool_call.type,
                                        "function": {
                                            "name": tool_call.function.name,
                                            "arguments": tool_call.function.arguments,
                                            "thought": thought,
                                        },
                                    }
                                    for tool_call in (message.tool_calls or [])
                                ],
                            }
                        }
                    },
                )
            )
        elif message.role == "system":
            messages.append(
                SystemMessage(
                    content=message.content,  # type: ignore
                )
            )
        elif message.role == "tool":
            if message.tool_call_id not in tool_calls:
                raise InvalidParameter(
                    parameter="tool_call_id",
                    cause=f"Invalid parameter: 'tool_call_id' of {message.tool_call_id}"
                    f" not found in 'tool_calls' of previous message",
                )
            next_tool_ids.remove(message.tool_call_id)

            tool_call = tool_calls[message.tool_call_id]

            assert tool_call.type == "function", TypeError(
                f"expect `function`, got {tool_call.type}"
            )

            messages.append(
                FunctionMessage(
                    content=message.content,  # type: ignore
                    name=tool_call.function.name,
                )
            )

    if len(next_tool_ids) > 0:
        raise InvalidParameter(
            parameter="messages",
            cause="An assistant message with 'tool_calls' must be followed"
            " by tool messages responding to each 'tool_call_id'.",
        )

    return messages


@task()
def format_maas_prompts(
    template: BaseChatPromptTemplate,
    chat_messages: List[MaasChatMessage],
    **kwargs: Any,
) -> List[MaasChatMessage]:
    messages = template.format_messages(
        messages=maas_chat_messages_to_messages(chat_messages), **kwargs
    )

    prompts = [
        MaasChatMessage(
            role=MessageRole2ChatRole.get(message.type),
            content=message.content if isinstance(message.content, str) else "",
            name=message.name,
        )
        for message in messages
    ]

    return prompts


@task()
def format_ark_prompts(
    template: BaseChatPromptTemplate,
    chat_messages: List[ArkMessage],
    **kwargs: Any,
) -> List[ArkMessage]:
    messages = template.format_messages(
        messages=_convert_ark_messages(chat_messages), **kwargs
    )

    prompts = [
        ArkMessage(
            role=_convert_message_role_to_ark_role(message.type),
            content=message.content if isinstance(message.content, str) else "",
        )
        for message in messages
    ]

    return prompts



import abc
import json
import logging
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Coroutine,
    Generic,
    List,
    Type,
    Union,
)

from pydantic import BaseModel, ValidationError
from volcenginesdkarkruntime._exceptions import ArkAPIError

import ark.core.utils.errorsv3 as errors_v3
from ark.core.idl.common_protocol import ArkError, RequestType, Response, ResponseType
from ark.core.idl.maas_protocol import MaasChatChoice, MaasChatRequest, MaasChatResponse
from ark.core.utils.context import get_reqid
from ark.core.utils.errors import GPTException, InternalServiceError
from ark.core.utils.types import parse_pydantic_error


class AsyncRunner(BaseModel, Generic[RequestType, ResponseType]):
    invoke: Callable[[RequestType], Coroutine[Any, Any, AsyncIterable[ResponseType]]]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
        **kwargs: Any,
    ):
        async def handler(
            request: RequestType,
        ) -> AsyncIterable[ResponseType]:
            return runnable_func(request)

        super().__init__(invoke=handler, **kwargs)

    @abc.abstractmethod
    async def arun(self, request: RequestType) -> ResponseType:
        pass

    @abc.abstractmethod
    def astream(self, request: RequestType) -> AsyncIterator[str]:
        pass


class ChatAsyncRunner(AsyncRunner[MaasChatRequest, MaasChatResponse]):
    async def arun(self, request: MaasChatRequest) -> MaasChatResponse:  # type: ignore
        verbose: List[List[MaasChatChoice]] = []
        responses: List[MaasChatResponse] = []
        try:
            async for resp in await self.invoke(request):  # type: MaasChatResponse
                if resp.choices and resp.choices[0].message:
                    responses.append(resp)
                else:
                    verbose.append(resp.choices)
        except GPTException as e:
            resp = MaasChatResponse(req_id=get_reqid(), error=e.to_error())
            responses.append(resp)
            logging.exception("assistant chat meet gpt error")
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = MaasChatResponse(req_id=get_reqid(), error=err.to_error())
            responses.append(resp)
            logging.exception("assistant chat meet internal error")

        response = MaasChatResponse.merge(responses=responses)
        return response

    async def astream(self, request: MaasChatRequest) -> AsyncIterator[str]:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: MaasChatResponse
                if resp.choices and resp.choices[0].message:
                    if isinstance(resp, BaseModel):
                        yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
                    else:
                        yield f"data:{json.dumps(resp, ensure_ascii=False)}\r\n\r\n"  # noqa E501
        except GPTException as e:
            resp = MaasChatResponse(req_id=get_reqid(), error=e.to_error())
            logging.exception("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = MaasChatResponse(req_id=get_reqid(), error=err.to_error())
            logging.exception("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"


class CustomAsyncRunner(AsyncRunner[RequestType, ResponseType]):
    response_cls: Type[ResponseType]

    def __init__(
        self,
        response_cls: Type[ResponseType],
        runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
        **kwargs: Any,
    ):
        super().__init__(
            response_cls=response_cls, runnable_func=runnable_func, **kwargs
        )

    async def arun(self, request: RequestType) -> ResponseType:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                return resp
        except GPTException as e:
            resp = self.response_cls(error=e.to_error())
            logging.error(f"bot meet error{e}")
            return resp
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = self.response_cls(error=err.to_error())
            logging.error(f"bot meet internal error{err}")
            return resp

    async def astream(self, request: RequestType) -> AsyncIterator[str]:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                if isinstance(resp, BaseModel):
                    yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
                else:
                    yield f"data:{json.dumps(resp, ensure_ascii=False)}\r\n\r\n"  # noqa E501
        except GPTException as e:
            resp = self.response_cls(error=e.to_error())
            logging.error("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = InternalServiceError(str(e))
            resp = self.response_cls(error=err.to_error())
            logging.error("stream chat meet error")
            yield f"data:{resp.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"


class ChatV3AsyncRunner(AsyncRunner[RequestType, ResponseType]):
    async def arun(self, request: RequestType) -> Union[Response, ResponseType]:  # type: ignore
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                return resp
        except errors_v3.APIException as e:
            logging.error(f"[API Error]: chat meet error:{e}")
            raise e
        except ValidationError as e:
            logging.error(f"[Validation Error]: chat meet parameter error:{e}")
            raise parse_pydantic_error(e)
        except ArkAPIError as e:
            logging.error(f"[Calling Chat Error]: chat meet error:{e}")
            raise e
        except Exception as e:
            err = errors_v3.InternalServiceError(str(e))
            logging.error(f"[Internal Error]: chat meet error:{e}")
            raise err

    async def astream(self, request: RequestType) -> AsyncIterator[str]:
        try:
            async for resp in await self.invoke(request):  # type: ResponseType
                yield f"data:{resp.model_dump_json(exclude_none=True)}\r\n\r\n"  # noqa E501
        except errors_v3.APIException as e:
            err = Response(error=e.to_error())
            logging.error(f"[API Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except ValidationError as e:
            err = Response(error=parse_pydantic_error(e).to_error())
            logging.error(f"[Validation Error]: stream chat meet parameter error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except ArkAPIError as e:
            err = Response(
                error=ArkError(
                    code=e.code or "",  # code is none when calling top failed
                    message=e.message or "",
                    type=e.type,
                    param=e.param,
                )
            )
            logging.error(f"[Calling Chat Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        except Exception as e:
            err = Response(error=errors_v3.InternalServiceError(str(e)).to_error())
            logging.error(f"[Internal Error]: stream chat meet error:{e}")
            yield f"data:{err.model_dump_json(exclude_unset=True, exclude_none=True)}\r\n\r\n"  # noqa E501
        yield "data:[DONE]\r\n\r\n"



import os
import shutil
import time
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from ark.cli.utils import TosSDKDownloader, VeFaaS, get_default_cr_host
from ark.cli.utils.vefaas import InvalidOperationError
from ark.core.launcher.local.serve import launch_serve

app = typer.Typer(no_args_is_help=True, add_completion=False)


@app.command()
def push(
    name: Annotated[str, typer.Argument(help="Project Name")],
    *,
    user_name: Annotated[
        str,
        typer.Option(
            help="Image Repository User name",
        ),
    ],
    password: Annotated[str, typer.Option(help="Image Repository Password")],
    ak: Annotated[
        str,
        typer.Option(help="Volcengine AccessKey"),
    ],
    sk: Annotated[str, typer.Option(help="Volcengine SecretKey")],
    release: Annotated[bool, typer.Option(help="Release the function")] = False,
    sdk_version: Annotated[
        Optional[str], typer.Option(help="SDK version for image to install")
    ] = None,
    cr_host: Annotated[
        Optional[str], typer.Option(help="Image Repository Host")
    ] = get_default_cr_host(),
    cr_namespace: Annotated[
        Optional[str], typer.Option(help="Image Repository Namespace")
    ] = None,
    image_name: Annotated[Optional[str], typer.Option(help="Image Name")] = None,
    image_version: Annotated[
        Optional[str], typer.Option(help="Image Tag Version")
    ] = "latest",
    bucket_name: Annotated[
        Optional[str], typer.Option(help="Tos Bucket Name")
    ] = "ark-sdk-cn-beijing",
    sdk_prefix: Annotated[
        Optional[str], typer.Option(help="Tos Sdk Prefix")
    ] = "python_sdk",
) -> None:
    try:
        import docker
    except ImportError as e:
        raise ImportError("Please install docker package to use this command.") from e

    assert isinstance(user_name, str), "please input your user name"
    assert isinstance(password, str), "please input your password"

    # image meta
    project_path = Path.cwd() / name
    cr_namespace, image_name = cr_namespace or name, image_name or name
    tag = f"{cr_host}/{cr_namespace}/{image_name}:{image_version}"
    platform = os.getenv("ARCH") or "linux/x86_64"

    # sdk arg
    tos_client = TosSDKDownloader(
        ak=ak, sk=sk, bucket_name=bucket_name, sdk_prefix=sdk_prefix
    )
    sdk_version = sdk_version or tos_client.get_latest_sdk_wheel()
    build_args = {
        "SDK_WHEEL": sdk_version,
        "SDK_URL": tos_client.get_sdk_presign_url(sdk_version),
    }

    # docker build image
    docker_client = docker.from_env()  # type: ignore
    build_logs = docker_client.api.build(
        path=project_path.__str__(),
        dockerfile=os.path.join(project_path, "Dockerfile"),
        tag=tag,
        platform=platform,
        buildargs=build_args,
    )
    for line in build_logs:
        typer.echo(line)

    typer.echo(f"Image: {tag} build successfully")

    # docker push image
    auth_config = {"username": user_name, "password": password}
    resp = docker_client.images.push(
        repository=tag, auth_config=auth_config, stream=True, decode=True
    )
    for line in resp:
        typer.echo(line)

    typer.echo(f"Image: {tag} pushed to repository successfully")

    # push function to vefaas
    try:
        vefaas_client = VeFaaS(ak=ak, sk=sk)

        function = vefaas_client.get_function_by_name(name)
        typer.echo(f"Get Function {name}, function id: {function.Id}")

        function = vefaas_client.update_function_image(function.Id, tag)
        typer.echo(f"Push function {name} successfully, function id: {function.Id}")
    except Exception as e:
        typer.echo(f"Push function {name} to vefaas failed: {e}")
        raise e

    if not release:
        return

    # release function on vefaas
    is_synced = False
    while not is_synced:
        try:
            resp = vefaas_client.release_function(function.Name, function.Id)
            is_synced = True
            typer.echo(f"Release function {name} successfully, status: {resp.Status}")
        except InvalidOperationError as e:
            assert e.is_invalid_image_status(), e
            typer.echo(f"Release function {name} failed, err:{e.message}, retrying...")
            time.sleep(10)
        except Exception as e:
            typer.echo(f"Release function {name} failed, err:{e}, stopped.")
            raise e


@app.command()
def create(name: Annotated[str, typer.Argument(help="Project Name")]) -> None:
    project_template_dir = Path(__file__).parents[0] / "serve_template"
    destination_dir = Path.cwd() / name

    # create ./{name}/code/ directory
    os.makedirs(destination_dir, exist_ok=True)
    shutil.copytree(project_template_dir, destination_dir, dirs_exist_ok=True)

    typer.echo("Initialization completed.")


@app.command()
def serve(name: Annotated[str, typer.Argument(help="Project Name")]) -> None:
    package_path: str = (Path.cwd() / name / "code/main").__str__()
    launch_serve(package_path)


if __name__ == "__main__":
    app()



import copy
import logging
import os
from typing import Any, Dict, List, Optional

import aiofiles
import yaml
from pydantic import BaseModel, Field

from ark.core.config.action_config import AtomActionConfig
from ark.core.config.endpoint_config import EndpointConfig
from ark.core.idl.ark_protocol import ArkGroupChatConfig
from ark.core.utils.common import LazyLoadSingleton, Singleton
from ark.core.utils.dict import dict_merge

CONFIG_CENTER_DIR = "CONFIG_CENTER_DIR"
CONFIG_CENTER_NAME = "CONFIG_CENTER_NAME"
ConfEnv = "CONF_ENV"


class Settings(BaseModel, Singleton):
    conf_dir: str
    conf_name: str

    def __init__(self) -> None:
        super().__init__(
            conf_dir=os.getenv(CONFIG_CENTER_DIR, "./"),
            conf_name=os.getenv(CONFIG_CENTER_NAME, "pipeline"),
        )


class PipelineConfig(BaseModel):
    tenant_id: Optional[str] = Field(default="")
    account_id: Optional[str] = Field(default="")
    utm_source: Optional[str] = Field(default="")
    atom_action_config: List[AtomActionConfig] = Field(default_factory=list)
    group_chat_config: Optional[ArkGroupChatConfig] = Field(
        default_factory=ArkGroupChatConfig
    )

    llm_config: Dict[str, EndpointConfig] = Field(default_factory=dict)
    action_config: Dict[str, Dict[str, Any]] = Field(default_factory=dict)

    def merge_from(self, other: Optional["PipelineConfig"] = None) -> "PipelineConfig":
        other_dict = other.__dict__ if other else {}
        merged_dict = copy.deepcopy(self.__dict__)
        return PipelineConfig(**dict_merge(merged_dict, other_dict))


class ConfigCenter(PipelineConfig, LazyLoadSingleton):
    def __init__(self, settings: Optional[Settings] = None, **kwargs: Any):
        if not kwargs:
            settings = Settings() if settings is None else settings
            conf_path = os.path.join(settings.conf_dir, f"{settings.conf_name}.yaml")

            with open(conf_path, mode="r") as f:
                kwargs = yaml.safe_load(f)
            logging.info(f"loading config:{kwargs}")
        super().__init__(**kwargs)

    @classmethod
    async def async_init(
        cls, settings: Optional[Settings] = None, **kwargs: Any
    ) -> "ConfigCenter":
        if not kwargs:
            settings = Settings() if settings is None else settings
            conf_path = os.path.join(settings.conf_dir, f"{settings.conf_name}.yaml")
            async with aiofiles.open(conf_path, mode="r") as f:
                kwargs = yaml.safe_load(await f.read())
            logging.info(f"loading config:{kwargs}")
        return cls(**kwargs)



import asyncio
import logging
import time
from typing import Any, Dict, List, Optional, Tuple, Union, get_args

import pydantic
from pydantic import BaseModel, field_validator, model_validator
from volcengine.viking_db import (
    Collection,
    Data,
    DistanceType,
    Field,
    FieldType,
    Index,
    IndexType,
    QuantType,
    Text,
    VectorIndexParams,
)

from ark.core.client import get_client_pool
from ark.core.client.vikingdb import VikingDBClient
from ark.core.rag import KnowledgeBase, KnowledgeChunk, KnowledgeSchema
from ark.core.utils.errors import InvalidParameter, MissingParameter

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)

INSTANCE_TYPE = {
    "int64": int,
    "float32": float,
    "string": str,
    "bool": bool,
    "text": str,
}

LIST_TYPE = {
    "list<string>": List[str],
    "list<int64>": List[int],
    "vector": List[Any],
}

FIELD_ENUM_MAP = {
    "int64": FieldType.Int64,
    "string": FieldType.String,
    "text": FieldType.Text,
    "bool": FieldType.Bool,
    "list<string>": FieldType.List_String,
    "list<int64>": FieldType.List_Int64,
    "vector": FieldType.Vector,
}


def _get_vdb_client() -> VikingDBClient:
    client_pool = get_client_pool()
    client: VikingDBClient = client_pool.get_client("vdb")  # type: ignore
    if not client:
        client = VikingDBClient()
    return client


def _default_index_name(collection_name: str) -> str:
    return f"{collection_name}_Index"


def _default_index(
    client: VikingDBClient, collection_name: str, custom: Dict[str, Any] = {}
) -> Index:
    return Index(
        collection_name=collection_name,
        index_name=_default_index_name(collection_name),
        cpu_quota=custom.get("cpu_quota", 2),
        partition_by=custom.get("partition_by", ""),
        vector_index=custom.get(
            "vector_index",
            VectorIndexParams(
                distance=DistanceType.IP,
                index_type=IndexType.HNSW,
                quant=QuantType.Int8,
            ),
        ),
        scalar_index=custom.get("scalar_index", None),
        shard_count=custom.get("shard_count", None),
        viking_db_service=client,
        stat=None,
    )


def _default_fields() -> List[Field]:
    return [
        Field(
            field_name="kb_inner_success", field_type=FieldType.Bool, default_val=False
        ),
        Field(
            field_name="kb_inner_enable", field_type=FieldType.Bool, default_val=True
        ),
        Field(field_name="kb_inner_attr", field_type=FieldType.Text, default_val=""),
    ]


def _is_field_type_matched(obj: Any, expected_type: str) -> bool:
    field_type = INSTANCE_TYPE.get(expected_type)
    if field_type:
        if expected_type == "text":
            return isinstance(obj, dict) or isinstance(obj, field_type)
            # vdb lagacy
        return isinstance(obj, field_type)

    list_type = LIST_TYPE.get(expected_type)
    item_type = get_args(list_type)

    is_list_type_matched = isinstance(obj, list) and all(
        isinstance(item, item_type) for item in obj
    )
    return is_list_type_matched


def _is_valid_field_type(field_type: str) -> bool:
    for field in FieldType:
        if field_type == field.value:
            return True
    return False


class PrimaryKey(BaseModel):
    name: str
    type: str

    @field_validator("type")
    @classmethod
    def validate_type(cls, field_type: str) -> str:
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )
        return field_type


class Vector(BaseModel):
    name: str
    model: str
    type: str
    embedding_type: Optional[str] = pydantic.Field(default="llm")
    dim: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, dim, model, embedding_type = (
            v.get("type"),
            v.get("dim"),
            v.get("model"),
            v.get("embedding_type", "llm"),
        )

        assert field_type, MissingParameter("type")
        assert model, MissingParameter("model")
        assert embedding_type in ("vdb", "llm"), InvalidParameter("embedding_type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if dim:
            assert 4 <= dim <= 2048, InvalidParameter("dim should be in range [4,2048]")
            assert dim % 4 == 0, InvalidParameter("dim should be dived by 4")

        return v


class Scalar(BaseModel):
    name: str
    type: str
    default_val: Optional[Any] = None

    @model_validator(mode="before")
    @classmethod
    def validate_scalar(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, default_val = v.get("type"), v.get("default_val")

        assert field_type, MissingParameter("type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if default_val:
            assert _is_field_type_matched(default_val, field_type), InvalidParameter(
                "default_val"
            )
        return v


class VikingDBSchema(KnowledgeSchema):
    """
    This class describe vikingDB index structure
    """

    @field_validator("primary_key")
    @classmethod
    def validate_primary_key(cls, v: Dict[str, str]) -> Dict[str, str]:
        try:
            PrimaryKey.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("vector")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        try:
            Vector.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("scalars")
    @classmethod
    def validate_scalars(cls, v: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        try:
            for scalar in v:
                Scalar.model_validate(scalar)

            return v
        except Exception as e:
            raise e

    def schema_to_field(self, data: Dict[str, Any]) -> Field:
        field_name = data.get("name")
        return Field(
            field_name=field_name,
            field_type=FIELD_ENUM_MAP.get(data.get("type", "")),
            default_val=data.get("default_val"),
            dim=data.get("dim"),
            is_primary_key=(field_name == self.primary_key.get("name")),
            pipeline_name=data.get("model")
            if data.get("embedding_type", "llm") == "vdb"
            else None,
        )

    @property
    def fields(self) -> List[Field]:
        fields: List[Field] = _default_fields()

        fields.append(self.schema_to_field(self.primary_key))
        fields.append(self.schema_to_field(self.vector))

        for scalar in self.scalars:
            fields.append(self.schema_to_field(scalar))

        return fields

    def get_primary_key_data(self, field: Dict[str, Any]) -> Tuple[str, Any]:
        key_name, key_type = (
            self.primary_key.get("name"),
            self.primary_key.get("type"),
        )

        assert key_name, MissingParameter("name")
        assert key_type, MissingParameter("type")

        primary_key = field.get(key_name)
        assert primary_key, "primary key should not be empty"
        assert _is_field_type_matched(primary_key, key_type), InvalidParameter(
            f"primary_key should be type {key_type}"
        )

        return key_name, primary_key

    def get_vector_data(
        self, field: Dict[str, Any]
    ) -> Tuple[str, Union[List[Any], str]]:
        vector_name, vector_type, embedding_type = (
            self.vector.get("name"),
            self.vector.get("type"),
            self.vector.get("embedding_type"),
        )
        assert vector_name, MissingParameter("vector name")
        assert vector_type, MissingParameter("vector type")

        if embedding_type == "vdb":
            vector = field.get(vector_name, "")
            assert _is_field_type_matched(vector, vector_type), InvalidParameter(
                f"vector should be type {vector_type}"
            )

            return vector_name, vector
        # embedding_type == "llm", viking search do not provide vector field
        return vector_name, ""

    def get_scalar_data(self, field: Dict[str, Any]) -> Dict[str, Any]:
        scalar_dic: Dict[str, Any] = {}
        for scalar_schema in self.scalars:
            scalar_name, scalar_type, default_val = (
                scalar_schema.get("name"),
                scalar_schema.get("type"),
                scalar_schema.get("default_val"),
            )
            assert scalar_name, MissingParameter("scalar name")
            assert scalar_type, MissingParameter("scalar type")

            scalar = field.get(scalar_name, default_val)
            if scalar is None:
                scalar = default_val

            assert _is_field_type_matched(scalar, scalar_type), InvalidParameter(
                f"{scalar_name} should be type {scalar_type}, "
                f"type of {scalar_name} is:{type(scalar)}"
            )

            scalar_dic[scalar_name] = scalar

        assert len(scalar_dic) > 0, "scalar should not be empty"

        return scalar_dic


class VikingDB(KnowledgeBase):
    """
    This class is used to store index data into vdb
    """

    collection: Collection
    index: Index
    client: VikingDBClient = pydantic.Field(default_factory=_get_vdb_client)

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def fields_to_chunk(self, fields: Dict[str, Any]) -> KnowledgeChunk:
        return KnowledgeChunk(
            knowledge_schema=self.knowledge_schema,
            primary_key=self.knowledge_schema.get_primary_key_data(fields),
            vector=self.knowledge_schema.get_vector_data(fields),
            scalars=self.knowledge_schema.get_scalar_data(fields),
        )

    @classmethod
    def chunk_to_data(cls, chunk: KnowledgeChunk, ttl: Optional[int] = None) -> Data:
        fields: Dict[str, Any] = {
            chunk.primary_key[0]: chunk.primary_key[1],
            chunk.vector[0]: chunk.vector[1],
        }

        for scalar in chunk.scalars.items():
            fields[scalar[0]] = scalar[1]

        fields["kb_inner_success"] = fields.get("kb_inner_success", True)
        fields["kb_inner_enable"] = fields.get("kb_inner_enable", True)
        fields["kb_inner_attr"] = fields.get("kb_inner_attr", "")
        return Data(fields=fields, TTL=ttl)

    @classmethod
    def generate_sid(cls) -> str:
        import random
        import string
        from datetime import datetime

        TIMESTAMP_FORMAT = "%Y%m%d%H%M%S"
        RANDOM_LENGTH = 5

        timestamp = datetime.now().strftime(TIMESTAMP_FORMAT)
        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=RANDOM_LENGTH)
        )
        return f"vdb_{timestamp}_{random_string}"

    @classmethod
    def create(
        cls,
        schema: VikingDBSchema,
        description: Optional[str] = "",
        index: Dict[str, Any] = {},
        client: Optional[VikingDBClient] = None,
    ) -> "VikingDB":
        vdb_client: VikingDBClient = client or _get_vdb_client()

        collection_name: str = cls.generate_sid()
        fields = schema.fields

        collection: Collection = vdb_client.create_collection(
            collection_name, fields, description=description
        )

        new_index = _default_index(
            client=vdb_client, collection_name=collection_name, custom=index
        )

        vdb_index: Index = vdb_client.create_index(
            collection_name=collection_name,
            index_name=new_index.index_name,
            vector_index=new_index.vector_index,
            cpu_quota=new_index.cpu_quota,
            description=new_index.description,
            partition_by=new_index.partition_by,
            scalar_index=new_index.scalar_index,
            shard_count=new_index.shard_count,
        )

        return cls(
            sid=collection_name,
            knowledge_schema=schema,
            collection=collection,
            client=vdb_client,
            index=vdb_index,
        )

    @classmethod
    def delete(cls, sid: str, client: Optional[VikingDBClient] = None) -> None:
        """
        Delete the knowledge collection
        """
        vdb_client: VikingDBClient = client or _get_vdb_client()
        return vdb_client.drop_collection(sid)

    @classmethod
    def get(
        cls,
        sid: str,
        schema: VikingDBSchema,
        index_name: Optional[str] = None,
        client: Optional[VikingDBClient] = None,
    ) -> "VikingDB":
        vdb_client: VikingDBClient = client or _get_vdb_client()
        collection = vdb_client.get_collection(sid)

        index = vdb_client.get_index(sid, index_name or _default_index_name(sid))

        return cls(
            sid=sid,
            knowledge_schema=schema,
            collection=collection,
            client=vdb_client,
            index=index,
        )

    def upsert_chunks(
        self, data: List[KnowledgeChunk], ttl: Optional[int] = None
    ) -> bool:
        """
        Add data to the knowledge base
        """
        data_list: List[Data] = []
        for d in data:
            data_list.append(self.chunk_to_data(d, ttl))

        self.collection.upsert_data(data=data_list)
        return True

    def flush(
        self,
        chunks: List[KnowledgeChunk],
        bs: int = 50,
        retry: int = 3,
        ttl: Optional[int] = None,
    ) -> None:
        """
        Batch flush chunk into knowledge base
        """
        # TODO: 需要事务性支持，需要VDB实现批量刷新接口
        for i in range(0, len(chunks), bs):
            succ = False
            for r in range(retry):
                try:
                    succ = self.upsert_chunks(chunks[i : i + bs], ttl)
                except Exception as e:
                    LOGGER.error("upsert_data error, retry: %d, err: %s", r, e)
                    time.sleep(1)
                    continue
                break
            if not succ:
                raise Exception("upsert_data error after retry")
                # FIXME: raise proper Exception

    def get_chunk(self, primary_key: str) -> KnowledgeChunk:
        """
        Get a chunk from the knowledge base
        """

        data: Data = self.collection.fetch_data(primary_key)

        return self.fields_to_chunk(data.fields)

    def delete_chunks(self, primary_key: Union[str, List[str]]) -> None:
        """
        Delete chunk from the knowledge base
        """

        self.collection.delete_data(primary_key)

    def batch_delete_chunks(
        self, chunk_id: List[str], bs: int = 50, retry: int = 3
    ) -> None:
        """
        Batch delete chunk from knowledge base
        """
        # TODO: 需要事务性支持，需要VDB实现批量删除接口
        for i in range(int(len(chunk_id) / bs) + 1):
            for r in range(retry):
                try:
                    if (i + 1) * bs <= len(chunk_id):
                        self.delete_chunks(chunk_id[i * bs : (i + 1) * bs])
                    elif i * bs < len(chunk_id):
                        self.delete_chunks(chunk_id[i * bs :])
                    else:
                        break
                except Exception as e:
                    LOGGER.error("upsert_data error, retry: %d, err: %s", r, e)
                    time.sleep(0.1)
                    continue
                break

    def search(
        self, query: Union[str, List[Any]], count: int, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        """
        Search the knowledge base
        """

        if isinstance(query, list):
            datas: List[Data] = self.index.search_by_vector(
                vector=query, limit=count, **kwargs
            )
        elif isinstance(query, str):
            datas = self.index.search_by_text(
                text=Text(text=query), limit=count, **kwargs
            )
        else:
            raise InvalidParameter("query should be either str or list")
        chunks: List[KnowledgeChunk] = []
        for data in datas:
            chunk = self.fields_to_chunk(data.fields)
            chunk.retrieve_score = data.score
            chunks.append(chunk)

        return chunks

    async def aupsert_chunks(self, data: List[KnowledgeChunk]) -> bool:
        """
        Add data to the knowledge base
        TODO: refactor with async vdb client
        """

        return await asyncio.get_running_loop().run_in_executor(
            None, self.upsert_chunks, data
        )

    async def aget_chunk(self, primary_key: str) -> KnowledgeChunk:
        """
        Get a chunk from the knowledge base
        TODO: refactor with async vdb client
        """
        return await asyncio.get_running_loop().run_in_executor(
            None, self.get_chunk, primary_key
        )

    async def adelete_chunks(self, primary_key: List[str]) -> None:
        """
        Delete chunkfrom the knowledge base
        TODO: refactor with async vdb client
        """
        return await asyncio.get_running_loop().run_in_executor(
            None, self.delete_chunks, primary_key
        )

    async def asearch(
        self, query: Union[str, List[Any]], count: int, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        """
        Search the knowledge base
        """

        if isinstance(query, list):
            datas: List[Data] = await self.index.async_search_by_vector(
                vector=query, limit=count, **kwargs
            )
        elif isinstance(query, str):
            datas = await self.index.async_search_by_text(
                text=Text(text=query), limit=count, **kwargs
            )
        else:
            raise InvalidParameter("query should be either str or list")
        chunks: List[KnowledgeChunk] = []
        for data in datas:
            chunk = self.fields_to_chunk(data.fields)
            chunk.retrieve_score = data.score
            chunks.append(chunk)

        return chunks



import inspect
import time
from functools import wraps
from typing import Any, AsyncGenerator, Callable, Dict, Iterable, Optional, TypeVar

from opentelemetry import trace

from ark.core.telemetry.trace import set_trace_attributes
from ark.core.utils.asyncio_utils import aenumerate
from ark.core.utils.context import (
    get_account_id,
    get_client_reqid,
    get_reqid,
    get_resource_id,
    get_resource_type,
)

T = TypeVar("T", covariant=True)
tracer = trace.get_tracer(__name__)


def get_remote_func(func):  # type: ignore
    try:
        import ray

        return ray.remote(func)
    except ImportError:
        raise ModuleNotFoundError(
            "Could not import ray python package. "
            "Please install it with `pip install ray`."
        )


def _update_kwargs(args: Any, kwargs: Any, func: Callable) -> Dict[Any, Any]:
    if kwargs is None:
        kwargs = {}
    if args is None:
        return kwargs
    try:
        return {**kwargs, **dict(zip(func.__code__.co_varnames, args))}
    except Exception:
        return {"args": args}


def task(
    distributed: bool = False,
    watch_io: bool = True,
    trace_all: bool = True,
    custom_attributes: Optional[Dict[str, Any]] = None,
) -> Any:
    def task_wrapper(func):  # type: ignore
        async def async_exec(*args: Any, **kwargs: Any) -> Any:
            with tracer.start_as_current_span(name=func.__qualname__) as span:
                input = _update_kwargs(args, kwargs, func)
                try:
                    result = await (get_remote_func(func) if distributed else func)(
                        *args, **kwargs
                    )
                    set_trace_attributes(
                        span,
                        status_code=trace.StatusCode.OK,
                        input=input if watch_io else "",
                        output=result if watch_io else "",
                        resource_type=get_resource_type(),
                        resource_id=get_resource_id(),
                        request_id=get_reqid(),
                        client_request_id=get_client_reqid(),
                        account_id=get_account_id(),
                        custom_attributes=custom_attributes,
                    )
                    return result
                except Exception as e:
                    handle_exception(span, e, input)
                    raise e

        def sync_exec(*args: Any, **kwargs: Any) -> Any:
            with tracer.start_as_current_span(name=func.__qualname__) as span:
                input = _update_kwargs(args, kwargs, func)
                try:
                    result = func(*args, **kwargs)
                    set_trace_attributes(
                        span,
                        status_code=trace.StatusCode.OK,
                        input=input if watch_io else "",
                        output=result if watch_io else "",
                        resource_type=get_resource_type(),
                        resource_id=get_resource_id(),
                        request_id=get_reqid(),
                        client_request_id=get_client_reqid(),
                        account_id=get_account_id(),
                        custom_attributes=custom_attributes,
                    )
                    return result
                except Exception as e:
                    handle_exception(span, e, input)
                    raise e

        @wraps(func)
        async def async_iter_task(*args: Any, **kwargs: Any) -> AsyncGenerator[T, None]:
            span = tracer.start_span(
                name=func.__qualname__ + ".first_iter", start_time=time.time_ns()
            )
            input = _update_kwargs(args, kwargs, func)
            try:
                async for i, resp in aenumerate(func(*args, **kwargs)):  # type: ignore
                    if i == 0 or trace_all:
                        set_trace_attributes(
                            span,
                            status_code=trace.StatusCode.OK,
                            input=input if (watch_io and i == 0) else "",
                            output=resp if watch_io else "",
                            resource_type=get_resource_type(),
                            resource_id=get_resource_id(),
                            request_id=get_reqid(),
                            client_request_id=get_client_reqid(),
                            account_id=get_account_id(),
                            custom_attributes=custom_attributes,
                        )
                        span.end(end_time=time.time_ns())
                    yield resp

                    if trace_all:
                        span = tracer.start_span(
                            name=func.__qualname__, start_time=time.time_ns()
                        )
            except Exception as e:
                if not trace_all:
                    span = tracer.start_span(
                        name=func.__qualname__, start_time=time.time_ns()
                    )
                handle_exception(span, e, input)
                raise e
            finally:
                span.end(end_time=time.time_ns())

        @wraps(func)
        def iter_task(*args: Any, **kwargs: Any) -> Iterable[T]:
            span = tracer.start_span(
                name=func.__qualname__ + ".first_iter", start_time=time.time_ns()
            )
            input = _update_kwargs(args, kwargs, func)
            try:
                for i, resp in enumerate(func(*args, **kwargs)):
                    if i == 0 or trace_all:
                        set_trace_attributes(
                            span,
                            status_code=trace.StatusCode.OK,
                            input=input if (watch_io and i == 0) else "",
                            output=resp if watch_io else "",
                            resource_type=get_resource_type(),
                            resource_id=get_resource_id(),
                            request_id=get_reqid(),
                            client_request_id=get_client_reqid(),
                            account_id=get_account_id(),
                            custom_attributes=custom_attributes,
                        )
                        span.end(end_time=time.time_ns())
                    yield resp
                    if trace_all:
                        span = tracer.start_span(
                            name=func.__qualname__, start_time=time.time_ns()
                        )
            except Exception as e:
                if not trace_all:
                    span = tracer.start_span(
                        name=func.__qualname__, start_time=time.time_ns()
                    )
                handle_exception(span, e, input)
                raise e
            finally:
                span.end(end_time=time.time_ns())

        if inspect.isasyncgenfunction(func):
            return async_iter_task
        elif inspect.isgeneratorfunction(func):
            return iter_task
        elif inspect.iscoroutinefunction(func) or distributed:
            return async_exec
        else:
            return sync_exec

    return task_wrapper


def handle_exception(span: trace.Span, exception: Exception, args: Any) -> None:
    set_trace_attributes(
        span,
        status_code=trace.StatusCode.ERROR,
        input=args,
        output="",
        resource_type=get_resource_type(),
        resource_id=get_resource_id(),
        request_id=get_reqid(),
        account_id=get_account_id(),
        client_request_id=get_client_reqid(),
    )
    span.record_exception(exception)



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from ark.core.idl.ark_protocol import (
    ArkMessage,
    ArkSearchSourceType,
    FunctionDefinition,
    SourceType,
    UserInfoExtra,
)
from ark.core.idl.maas_protocol import MaasChatMessage
from ark.core.task.task import task


class AtomActionConfig(BaseModel):
    action_name: str
    function_definition: FunctionDefinition


class ActionConfig(BaseModel):
    action_name: str = Field(default="")

    @classmethod
    def convert_messages_to_strs(
        cls, messages: Optional[List[Union[ArkMessage, MaasChatMessage, str]]]
    ) -> Optional[List[str]]:
        if not messages:
            return None
        content = []
        for message in messages:
            if isinstance(message, str):
                content.append(message)
            elif isinstance(message.content, str):
                content.append(message.content)

        return content

    def merge_dict(self, other: Dict[str, Any]) -> Dict[str, Any]:
        merged_dict = self.dict()

        for key in merged_dict.keys():
            value = other.get(key, None)
            if value is not None:
                merged_dict[key] = value

        return merged_dict


class SearchIntentionConfig(ActionConfig):
    keywords: Optional[List[str]] = None
    result_mapping: Optional[Dict[str, bool]] = None


class MultiIntentionConfig(SearchIntentionConfig):
    source_type: Optional[List[SourceType]] = None


class SearchSummaryConfig(ActionConfig):
    summary_top_k: Optional[int] = None
    keywords: Optional[List[str]] = None
    user_info: Optional[UserInfoExtra] = None
    biz_id: Literal["seed", "volc"] = "seed"
    extra_params: Optional[Dict[str, Any]] = None
    ab_params: Optional[Dict[str, Any]] = None
    summary_max_total_len: Optional[int] = None
    browsing_doc_list: str = ""
    summary_with_system_prompt: bool = False
    site: str = ""
    allow_host: Optional[List[str]] = None
    block_host: Optional[List[str]] = None
    publish_time: Dict[str, str] = Field(default_factory=dict)


@task()
def get_extended_source_types(search_config: MultiIntentionConfig) -> List[SourceType]:
    source_type_list: List[SourceType] = []
    if not search_config.source_type:
        return source_type_list

    for source_type in search_config.source_type:
        if source_type == ArkSearchSourceType.DEFAULT_SEARCH:
            continue
        source_type_list.append(source_type)
    return source_type_list



from enum import Enum
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel
from volcenginesdkarkruntime.types.create_tokenization_response import Tokenization

from ark.core.idl.maas_protocol import Reference
from ark.core.task.task import task
from ark.core.utils.types import snake_merge


class ReferenceType(str, Enum):
    SEARCH = "search"
    RAG = "rag"


ReferenceType2TagMap = {ReferenceType.SEARCH: "联网", ReferenceType.RAG: "知识库"}


class ReferenceOrganizationInput(BaseModel):
    reference_type: ReferenceType
    references: List[Reference]
    reference_texts: List[str]
    offset_mappings: List[List[List[int]]]
    total_tokens: List[int]


class ReferenceText(BaseModel):
    """
    ReferenceText is used for reference text parsing
    """

    text: str
    """
    text stands for reference basic content
    """
    tokenization: Tokenization
    """
    tokenization is used for content length limitation
    """
    reference_type: ReferenceType
    """
    reference_type stands for tagging
    """
    reference: Reference
    """
    reference stands for metadata
    """


@task()
def parse_multiple_reference_texts(
    reference_text_list: List[List[ReferenceText]],
    max_reference_tokens: int = 6000,
    min_reference_tokens: int = 0,
) -> str:
    filtered_references: Dict[ReferenceType, List[str]] = {}

    snake_ordered_list: List[ReferenceText] = snake_merge(reference_text_list)
    for item in snake_ordered_list:  # type: ReferenceText
        prefix_length, reference_tag = (
            0,
            ReferenceType2TagMap.get(item.reference_type, ""),
        )

        item_index = len(filtered_references.get(item.reference_type, []))
        # prefix tag for calculating tokens number, not in attributes
        if item_index == 0:
            prefix_length = len(f"\n## {reference_tag}资料\n")

        # prompt begin
        attrs = f"### 资料{item_index + 1}\n"

        # travel attributes fields with not-none value
        ordered_fields_set = sorted(item.reference.model_fields_set)
        for key in ordered_fields_set:
            field_info, value = (
                item.reference.model_fields[key],
                item.reference.__dict__.get(key),
            )
            if not field_info.description or not value:
                continue
            attrs += f"- {field_info.description}：{value}\n"

        attrs += f"- {reference_tag}资料正文：\n"

        # pre-check length
        max_reference_tokens -= prefix_length + len(attrs)
        if max_reference_tokens < min_reference_tokens:
            break

        # get text content by offset_mapping & max tokens limitation
        text = _get_reference_text(
            item.text,
            item.tokenization.offset_mapping,
            item.tokenization.total_tokens,
            max_reference_tokens,
        )
        # append item & reduce the tokens number
        filtered_references.setdefault(item.reference_type, []).append(attrs + text)
        max_reference_tokens -= item.tokenization.total_tokens

    # final encoding
    reference_text = ""
    for reference_type, text_list in filtered_references.items():
        reference_tag = ReferenceType2TagMap.get(reference_type, "")
        reference_text += f"\n## {reference_tag}资料\n"

        reference_text += "\n".join(text_list)

    return reference_text


def _get_reference_text(
    text: str,
    offset_mapping: List[List[int]],
    total_tokens: int,
    max_reference_tokens: int,
) -> str:
    reference_text_tokens = (
        offset_mapping
        if max_reference_tokens > total_tokens
        else offset_mapping[:max_reference_tokens]
    )
    index_begin, index_end = 0, -1
    if reference_text_tokens and len(reference_text_tokens) > 0:
        if len(reference_text_tokens[0]) > 0:
            index_begin = reference_text_tokens[0][0]
        if len(reference_text_tokens[-1]) > 1:
            index_end = reference_text_tokens[-1][1]

    return text[index_begin:index_end]


@task()
async def parse_reference(
    tokens: List[List[str]],
    total_tokens: List[int],
    references: List[Reference],
    max_reference_token: int = 6000,
    min_reference_token: int = 200,
) -> str:
    # lower bound for reference trim
    prefix_suffix_tokens = 12
    # tokens=['\n', '<', '资料', '{i+1}', '开始', '>\n']
    reference_text = ""
    for i, (token, total_token) in enumerate(zip(tokens, total_tokens)):
        max_reference_token -= prefix_suffix_tokens
        if max_reference_token <= min_reference_token:
            break
        reference_text += f"\n<资料{i + 1}开始>\n"
        attrs = ""
        if i < len(references):
            if references[i].title:
                attrs += f"标题：{references[i].title}\n"
            if references[i].site_name:
                attrs += f"来源：{references[i].site_name}\n"
            if references[i].freshness_info:
                attrs += f"时效性：{references[i].freshness_info}\n"
            if references[i].publish_time:
                attrs += f"发布时间：{references[i].publish_time}\n"
        attrs += "资料正文：\n"
        reference_text += attrs
        max_reference_token -= len(attrs)  # attrs 直接用 len 当做 token len
        reference_text += (
            "".join(token)
            if max_reference_token > total_token
            else "".join(token[:max_reference_token])
        )
        reference_text += f"\n<资料{i + 1}结束>\n"
        max_reference_token -= total_token
    return reference_text


@task()
async def parse_reference_with_offset_mapping(
    origin_texts: List[str],
    offset_mappings: List[List[List[int]]],
    total_tokens: List[int],
    references: List[Reference],
    max_reference_token: int = 6000,
    min_reference_token: int = 200,
    reference_type: Optional[ReferenceType] = None,
) -> str:
    if (
        len(origin_texts) == 0
        or len(offset_mappings) == 0
        or len(total_tokens) == 0
        or len(references) == 0
    ):
        return ""

    # lower bound for reference trim
    prefix_suffix_tokens = 12
    # tokens=['\n', '<', '资料', '{i+1}', '开始', '>\n']
    reference_text, reference_tag = "", ""
    if reference_type is not None:
        reference_tag = ReferenceType2TagMap.get(reference_type, "")
        reference_text += f"\n## {reference_tag}资料"
    for i, (offset_mapping, total_token) in enumerate(
        zip(offset_mappings, total_tokens)
    ):
        max_reference_token -= prefix_suffix_tokens
        if max_reference_token <= min_reference_token:
            break

        if reference_type is None:
            reference_text += f"\n<资料{i + 1}开始>\n"
        else:
            reference_text += f"\n### 资料{i + 1}\n"

        attrs = ""
        if i < len(references):
            if references[i].title:
                attrs += f"- 标题：{references[i].title}\n"
            if references[i].site_name:
                attrs += f"- 来源：{references[i].site_name}\n"
            if references[i].freshness_info:
                attrs += f"- 时效性：{references[i].freshness_info}\n"
            if references[i].publish_time:
                attrs += f"- 发布时间：{references[i].publish_time}\n"
            if references[i].doc_id:
                attrs += f"- 文档ID：{references[i].doc_id}\n"
            if references[i].doc_name:
                attrs += f"- 文档名：{references[i].doc_name}\n"
            if references[i].doc_type:
                attrs += f"- 文档类型：{references[i].doc_type}\n"
            if references[i].doc_title:
                attrs += f"- 文档标题：{references[i].doc_title}\n"
            if references[i].chunk_title:
                attrs += f"- 分块标题：{references[i].chunk_title}\n"
            if references[i].chunk_id:
                attrs += f"- 分块编号：{references[i].chunk_id}\n"

        attrs += f"- {reference_tag}资料正文：\n"
        reference_text += attrs
        max_reference_token -= len(attrs)  # attrs 直接用 len 当做 token len

        reference_text += _get_reference_text(
            origin_texts[i], offset_mapping, total_token, max_reference_token
        )

        # compatible with older version
        if not reference_type:
            reference_text += f"\n<资料{i + 1}结束>\n"
        max_reference_token -= total_token
    return reference_text


@task()
def merge_reference_pair(
    reference_pair: Tuple[List[Reference], List[Reference]],
) -> List[Reference]:
    reference_first, reference_second = reference_pair
    merged_reference_list: List[Reference] = []
    for i in range(len(reference_first)):
        merged_dict = {
            **reference_first[i].model_dump(exclude_none=True, exclude_unset=True),
            **reference_second[i].model_dump(exclude_none=True, exclude_unset=True),
        }
        merged_reference = Reference(**merged_dict)
        merged_reference_list.append(merged_reference)
    return merged_reference_list



from __future__ import annotations

from typing import Any, AsyncIterable, Dict, Optional, Type, Union

import httpx
from httpx import URL, AsyncClient, Timeout
from volcengine.ApiInfo import ApiInfo
from volcenginesdkarkruntime import AsyncArk
from volcenginesdkarkruntime._base_client import make_request_options
from volcenginesdkarkruntime._constants import (
    BASE_URL,
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
)
from volcenginesdkarkruntime._resource import AsyncAPIResource
from volcenginesdkarkruntime._streaming import AsyncStream

from ark.core.client.authorization import ArkAuthorization
from ark.core.client.base import Client
from ark.core.idl.ark_protocol import ArkActionMeta, ArkToolMeta
from ark.core.idl.common_protocol import RequestType, ResponseType
from ark.core.utils.context import get_extra_headers
from ark.core.utils.errorsv3 import AuthenticationError, InvalidParameter


class AsyncTool(AsyncAPIResource):
    async def init(
        self,
        meta_info: Union[ArkActionMeta, ArkToolMeta],
    ) -> None:
        pass

    async def create(
        self,
        path: str,
        stream: bool,
        request: RequestType,
        response_cls: Type[ResponseType],
        extra_headers: Dict[str, str] | None = None,
        extra_query: Dict[str, object] | None = None,
        extra_body: Dict[str, object] | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> Dict[str, Any] | AsyncStream[Dict[str, Any]]:
        return await self._post(
            path,
            body=request.model_dump(exclude_unset=True, exclude_none=True),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=response_cls,
            stream=stream,
            stream_cls=AsyncStream[response_cls],
        )


class ArkActionClient(Client, AsyncArk):
    def __init__(
        self,
        *,
        ak: str | None = None,
        sk: str | None = None,
        api_key: str | None = None,
        base_url: str | URL = BASE_URL,
        timeout: float | Timeout | None = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        http_client: AsyncClient | None = None,
        auto_refresh_apikey: bool = True,
    ) -> None:
        super().__init__(
            ak=ak,
            sk=sk,
            api_key=api_key,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            http_client=http_client,
        )
        self.authorization = None
        # we only refresh api key when it is not set
        if auto_refresh_apikey and not self.api_key:
            self.authorization = ArkAuthorization(ak=ak, sk=sk)

        self.api_info = self.get_api_info()
        self.tool = AsyncTool(self)

    def _validate(self) -> None:
        if self.api_key is None and (self.sk is None or self.sk is None):
            raise AuthenticationError()

    def authorize(
        self,
        action_name: str,
        model: Optional[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, str]:
        self._validate()

        if self.authorization:
            if model:
                self.api_key = self.authorization.get_apikey(
                    resource_ids=[model], resource_type="endpoint"
                )
            else:
                self.api_key = self.authorization.get_apikey(
                    resource_ids=[action_name], resource_type="action"
                )

        if not extra_headers:
            extra_headers = {}
        extra_headers["Authorization"] = f"Bearer {self.api_key}"
        extra_headers = get_extra_headers(extra_headers)
        return extra_headers

    @staticmethod
    def get_api_info() -> Dict[str, ApiInfo]:
        return {
            "Search": ApiInfo("POST", "/actions/Search", {}, {}, {}),
            "SearchIntention": ApiInfo(
                "POST", "/actions/SearchIntention/chat/completions", {}, {}, {}
            ),
            "SearchSummary": ApiInfo(
                "POST", "/actions/SearchSummary/chat/completions", {}, {}, {}
            ),
            "Calculator": ApiInfo("POST", "/actions/Calculator", {}, {}, {}),
            "LinkReader": ApiInfo("POST", "/actions/LinkReader", {}, {}, {}),
            "ActionTool": ApiInfo("POST", "/tools/execute", {}, {}, {}),
        }

    async def arequest(
        self,
        api: str,
        meta_info: Union[ArkActionMeta, ArkToolMeta],
        request: RequestType,
        response_cls: Type[ResponseType],
        extra_headers: Dict[str, str] | None = None,
        extra_query: Dict[str, object] | None = None,
        extra_body: Dict[str, object] | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> ResponseType:
        await self.init(meta_info)
        api_info = self.api_info.get(api)
        if not api_info:
            raise InvalidParameter("api")

        extra_headers = self.authorize(
            meta_info.action_name, meta_info.model, extra_headers
        )

        response = await self.tool.create(
            api_info.path,
            stream=False,
            request=request,
            response_cls=response_cls,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )
        return response_cls.model_validate(response, strict=False)

    async def astream_request(
        self,
        api: str,
        meta_info: Union[ArkActionMeta, ArkToolMeta],
        request: RequestType,
        response_cls: Type[ResponseType],
        extra_headers: Dict[str, str] | None = None,
        extra_query: Dict[str, object] | None = None,
        extra_body: Dict[str, object] | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> AsyncIterable[ResponseType]:
        api_info = self.api_info.get(api)
        if not api_info:
            raise InvalidParameter("api")

        extra_headers = self.authorize(
            meta_info.action_name, meta_info.model, extra_headers
        )

        response: AsyncStream[Dict[str, Any]] = await self.tool.create(
            api_info.path,
            stream=True,
            request=request,
            response_cls=response_cls,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )
        async for resp in response:
            yield response_cls.model_validate(resp, strict=False)

    async def init(self, meta_info: Union[ArkActionMeta, ArkToolMeta]) -> None:
        await self.tool.init(meta_info=meta_info)



from typing import Any, Dict, Tuple, Type

from ark.core.client.action import ArkActionClient
from ark.core.client.authorization import APIGAuthorization, ArkAuthorization
from ark.core.client.base import Client, ClientPool, get_client_pool
from ark.core.client.maas import ArkClient
from ark.core.client.sse import AsyncSSEDecoder

__all__ = [
    "Client",
    "ClientPool",
    "ArkClient",
    "ArkActionClient",
    "ArkAuthorization",
    "APIGAuthorization",
    "AsyncSSEDecoder",
    "get_client_pool",
    "get_default_client_configs",
]


def get_default_client_configs() -> Dict[str, Tuple[Type[Client], Any]]:
    return {
        "chat": (
            ArkClient,
            {
                "host": "maas-api.ml-platform-cn-beijing.volces.com",
                "region": "cn-beijing",
            },
        ),
    }



from abc import ABC
from typing import Any, List, Union

from ark.component.retriever.base_retriever import BaseRetriever
from ark.component.v3.vectorstores.viking_knowledgebase import (
    VikingKnowledgeBase,
    VikingKnowledgeBaseSchema,
)
from ark.core.rag import KnowledgeChunk
from ark.core.utils.errors import InvalidParameter


class VikingKnowledgeBaseRetriever(BaseRetriever, ABC):
    def __init__(
        self, config: VikingKnowledgeBaseSchema, db: VikingKnowledgeBase
    ) -> None:
        self.config = config
        self.db = db

    async def retrieve(
        self,
        query: Union[str, List[float]],
        retrieve_count: int,
        *args: Any,
        **kwargs: Any,
    ) -> List[KnowledgeChunk]:
        dsl_filter = kwargs.get("dsl_filter")
        if isinstance(query, str):
            return await self.db.asearch(query, retrieve_count, filter=dsl_filter)
        else:
            raise InvalidParameter("viking_knowledgebase search query is str")



from abc import ABC, abstractmethod
from typing import Any, List

from ark.core.rag import KnowledgeChunk, KnowledgeDoc


class BaseReader(ABC):
    """
    This class is used to parse and chunk origin data
    """

    @abstractmethod
    def load(
        self, doc: KnowledgeDoc, *args: Any, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        """
        Load data into KnowledgeChunk objects.
        Returns: split chunks of the document
        """
        pass



import asyncio
import os
from typing import Any, Dict, List, Optional, Union

from volcengine.viking_knowledgebase import VikingKnowledgeBaseService
from volcengine.viking_knowledgebase.exception import VikingKnowledgeBaseServerException

import ark.core.utils.errorsv3 as errorsv3
from ark.component.retriever.base_retriever import BaseRetriever
from ark.component.vectorstores.vikingdb import VikingDBSchema
from ark.core.client import get_client_pool
from ark.core.idl.ark_protocol import ArkKnowledgeBaseRequest, ArkKnowledgeBaseResponse
from ark.core.idl.common_protocol import ChatRole, Reference
from ark.core.idl.maas_protocol import KnowledgeBaseResponse, MaasChatRequest
from ark.core.rag import KnowledgeChunk
from ark.core.task.task import task
from ark.core.utils.context import get_reqid
from ark.core.utils.errors import KnowledgeBaseError, MissingParameter

viking_knowledgebase_schema = VikingDBSchema(
    primary_key={
        "name": "id",
        "type": "string",
        "default_val": "",
    },
    vector={
        "name": "content",
        "type": "text",
        "model": "text_bge_large_zh",
        "embedding_type": "vdb",
    },
    scalars=[
        {
            "name": "doc_id",
            "type": "string",
            "default_val": "",
        },
        {
            "name": "doc_name",
            "type": "string",
            "default_val": "",
        },
        {
            "name": "doc_type",
            "type": "string",
            "default_val": "",
        },
        {
            "name": "doc_title",
            "type": "string",
            "default_val": "",
        },
        {
            "name": "chunk_title",
            "type": "string",
            "default_val": "",
        },
    ],
    vector_text_len_limit=500,
)


def _get_viking_knowledge_client() -> VikingKnowledgeBaseService:
    client_pool = get_client_pool()
    client: VikingKnowledgeBaseService = client_pool.get_client("viking-knowledgebase")  # type: ignore
    if not client:
        client = VikingKnowledgeBaseService(
            ak=os.getenv("VOLC_ACCESSKEY"), sk=os.getenv("VOLC_SECRETKEY")
        )
    return client


class VikingDBRetriever(BaseRetriever):
    def __init__(
        self,
        config: VikingDBSchema,
        kn_client: VikingKnowledgeBaseService,
        collection_name: str,
        project: str = "default",
    ) -> None:
        if not kn_client:
            kn_client = _get_viking_knowledge_client()
        self.config = config
        self.kn_client = kn_client
        self.collection_name = collection_name
        self.project = project

    @task()
    async def retrieve(
        self,
        query: Union[str, List[float]],
        retrieve_count: int,
        dsl_filter: Optional[Any] = None,
        partition: str = "default",
        norm: str = "none",
        vector_dim: int = 0,
    ) -> List[KnowledgeChunk]:
        if dsl_filter:
            db_filter = dsl_filter
        else:
            db_filter = None

        self.kn_client.setHeader({"x-tt-logid": get_reqid()})
        points = await self.kn_client.async_search_collection(
            self.collection_name,
            query,
            query_param=db_filter,
            limit=retrieve_count,
            project=self.project,
        )
        chunks: List[KnowledgeChunk] = []
        for point in points:
            fields = {
                "id": point.point_id,
                "chunk_title": point.chunk_title,
                "content": point.content,
                "doc_id": point.doc_info.doc_id,
                "doc_name": point.doc_info.doc_name,
                "doc_type": point.doc_info.doc_type,
                "doc_title": point.doc_info.doc_name,
            }
            chunk = KnowledgeChunk(
                knowledge_schema=self.config,
                primary_key=self.config.get_primary_key_data(fields),
                vector=self.config.get_vector_data(fields),
                scalars=self.config.get_scalar_data(fields),
            )
            chunks.append(chunk)
        return chunks


@task()
async def retrieve_knowledge(
    action_config: Dict[str, Any],
    req: MaasChatRequest,
    client: VikingKnowledgeBaseService,
    keywords: Optional[List[str]] = None,
) -> KnowledgeBaseResponse:
    collection_name, cnt, show_reference, project = (
        action_config.get("collection_name", ""),
        action_config.get("retrieve_count", 0),
        action_config.get("show_reference", True),
        action_config.get("project", "default"),
    )

    if not collection_name:
        raise MissingParameter("collection_name")
    if not cnt:
        raise MissingParameter("retrieve_count")

    if keywords and len(keywords) > 0:
        queries = keywords
    else:
        contents: List[str] = []
        for message in req.messages:
            if message.role == ChatRole.USER and isinstance(message.content, str):
                contents.append(message.content)
        if len(contents) > 0:
            queries = [contents[-1]]
        else:
            raise MissingParameter("query")

    retriever = VikingDBRetriever(
        config=viking_knowledgebase_schema,
        collection_name=collection_name,
        project=project,
        kn_client=client,
    )

    try:
        retrieve_results = await asyncio.gather(
            *[
                retriever.retrieve(
                    query=query[:1024], retrieve_count=cnt, dsl_filter=None
                )
                for query in queries
            ]
        )
    except VikingKnowledgeBaseServerException as e:
        raise KnowledgeBaseError(e.message)

    topk_chunks = [chunk for result in retrieve_results for chunk in result]
    # Format Reference
    refs: List[str] = []
    references: List[Reference] = []
    for chunk in topk_chunks:
        meta_text = (
            f"文档ID：{chunk.scalars['doc_id']}\n"
            + f"文档名：{chunk.scalars['doc_name']}\n"
            + f"文档类型：{chunk.scalars['doc_type']}\n"
            + f"文档标题：{chunk.scalars['doc_title']}\n"
            + f"分块标题：{chunk.scalars['chunk_title']}\n"
            + f"分块编号：{chunk.primary_key[1]}\n"
        )
        text = {chunk.vector[1]}
        refs.append(f"{meta_text}\n正文：\n{text}\n")

        if show_reference:
            references.append(
                Reference(
                    collection_name=collection_name,
                    chunk_id=chunk.primary_key[1],
                    chunk_title=chunk.scalars["chunk_title"],
                    doc_id=chunk.scalars["doc_id"],
                    doc_name=chunk.scalars["doc_name"],
                    doc_type=chunk.scalars["doc_type"],
                    doc_title=chunk.scalars["doc_title"],
                    project=project,
                )
            )

    return KnowledgeBaseResponse(
        texts=refs,
        references=references,
    )


@task()
async def retrieve_viking_knowledge(
    request: ArkKnowledgeBaseRequest,
    client: VikingKnowledgeBaseService,
) -> ArkKnowledgeBaseResponse:
    keywords = [request.question]
    if request.keywords and len(request.keywords) > 0:
        keywords = request.keywords

    retriever = VikingDBRetriever(
        config=viking_knowledgebase_schema,
        collection_name=request.collection_name,
        project=request.project,
        kn_client=client,
    )

    try:
        retrieve_results = await asyncio.gather(
            *[
                retriever.retrieve(
                    query=keyword[:1024],
                    retrieve_count=request.retrieve_count,
                    dsl_filter=None,
                )
                for keyword in keywords
            ]
        )
    except VikingKnowledgeBaseServerException as e:
        raise errorsv3.KnowledgeBaseError(e.message)

    # Format Reference
    texts: List[str] = []
    references: List[Reference] = []
    topk_chunks = [chunk for result in retrieve_results for chunk in result]
    for chunk in topk_chunks:
        texts.append(chunk.vector[1])
        if request.show_reference:
            references.append(
                Reference(
                    collection_name=request.collection_name,
                    chunk_id=chunk.primary_key[1],
                    chunk_title=chunk.scalars["chunk_title"],
                    doc_id=chunk.scalars["doc_id"],
                    doc_name=chunk.scalars["doc_name"],
                    doc_type=chunk.scalars["doc_type"],
                    doc_title=chunk.scalars["doc_title"],
                    project=request.project,
                )
            )

    return ArkKnowledgeBaseResponse(
        texts=texts,
        references=references,
    )



import json
import os
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Type

from fastapi import HTTPException
from volcenginesdkarkruntime import AsyncArk
from volcenginesdkarkruntime._constants import BASE_URL

from ark.core.client import Client, get_client_pool
from ark.core.client.action import ArkActionClient
from ark.core.client.maas import ArkClient
from ark.core.idl.common_protocol import Context, Request, RequestType, ResponseType
from ark.core.idl.maas_protocol import MaasChatRequest
from ark.core.launcher.vefaas.common import parse_request
from ark.core.runtime.sync import ChatSyncRunner, SyncRunner
from ark.core.utils.context import (
    set_account_id,
    set_resource_id,
    set_resource_type,
)
from ark.telemetry.trace import TraceConfig, setup_tracing


def setup_environment(
    trace_on: bool = True, trace_config: Optional[TraceConfig] = None
) -> None:
    set_resource_type(os.getenv("RESOURCE_TYPE") or "")
    set_resource_id(os.getenv("RESOURCE_ID") or "")
    set_account_id(os.getenv("ACCOUNT_ID") or "")

    setup_tracing(
        endpoint=os.getenv("TRACE_ENDPOINT"),
        trace_on=trace_on,
        trace_config=trace_config,
    )


def initialize(
    context: Any,
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
    trace_on: bool = True,
    trace_config: Optional[TraceConfig] = None,
) -> None:
    get_client_pool(
        clients
        or {
            # v3
            "ark": (
                AsyncArk,
                {
                    "base_url": BASE_URL,
                    "max_retries": 3,
                },
            ),
            # v2
            "chat": (
                ArkClient,
                {
                    "host": "maas-api.ml-platform-cn-beijing.volces.com",
                    "auto_refresh_apikey": True,
                },
            ),
            # action
            "action": (
                ArkActionClient,
                {
                    "base_url": BASE_URL,
                    "max_retries": 3,
                },
            ),
        }
    )

    setup_environment(trace_on, trace_config)


def handler(request: RequestType, runner: SyncRunner, context: Context) -> Any:
    try:
        from flask import Response
    except ImportError:
        raise

    try:
        if request.stream:
            return Response(
                runner.generate(request),
                200,
                {
                    "Content-Type": "text/event-stream; charset=utf-8",
                    "X-Tt-Logid": context.request_id,
                },
            )
        else:
            body: Any = runner.run(request)
            return {
                "body": json.dumps(
                    body,
                    ensure_ascii=False,
                    allow_nan=False,
                    indent=None,
                    separators=(",", ":"),
                ),
                "statusCode": 200,
                "headers": {
                    "X-Tt-Logid": context.request_id,
                    "Content-Type": "application/json",
                },
            }
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Invalid request: {str(e)}")


# https://www.volcengine.com/docs/6662/107430
def faas_handler(
    event: Dict[str, Any],
    context: Any,
    runnable_func: Callable[[RequestType], Iterable[ResponseType]],
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
    request_cls: Optional[Type[RequestType]] = None,
    runner_cls: Optional[Type[SyncRunner]] = None,
) -> Any:
    initialize(context, clients)

    runner = (
        runner_cls(runnable_func=runnable_func)  # type: ignore
        if runner_cls
        else ChatSyncRunner(runnable_func=runnable_func)  # type: ignore
    )
    request: Request = parse_request(event, request_cls or MaasChatRequest)

    return handler(request, runner, context)



from abc import ABC, abstractmethod
from typing import Any, AsyncIterator, Generic, Iterator, List, TypeVar

T = TypeVar("T")


class Runnable(Generic[T], ABC):
    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    async def arun(self, *args: Any, **kwargs: Any) -> Any:
        pass

    def stream(self, *args: Any, **kwargs: Any) -> Iterator[T]:
        raise NotImplementedError

    def astream(self, *args: Any, **kwargs: Any) -> AsyncIterator[T]:
        raise NotImplementedError

    def batch(self, *args: Any, **kwargs: Any) -> List[T]:
        raise NotImplementedError

    async def abatch(self, *args: Any, **kwargs: Any) -> List[T]:
        raise NotImplementedError



from typing import Any, Dict, Type

from pydantic import ValidationError

from ark.core.idl.common_protocol import Request, RequestType
from ark.core.task.task import task
from ark.core.utils.context import (
    get_client_reqid,
    get_reqid,
    set_client_reqid,
    set_headers,
    set_reqid,
)
from ark.core.utils.logid import gen_log_id
from ark.core.utils.types import parse_pydantic_error


@task()
def parse_request(event: Dict[str, Any], request_cls: Type[RequestType]) -> Request:
    body = event.get("body", "")
    headers = event.get("headers", {})

    set_headers(headers)

    # here we generate request id instead of reading from headers
    set_reqid(
        headers.get("X-Faas-Request-Id", headers.get("X-Request-Id", gen_log_id()))
    )
    set_client_reqid(headers.get("X-Client-Request-Id", get_reqid()))

    try:
        return request_cls.model_validate_json(body, strict=True)
    except ValidationError as e:
        raise parse_pydantic_error(e)


@task()
def parse_response(status_code: int, content: str) -> Dict[str, Any]:
    return {
        "statusCode": status_code,
        "headers": {
            "content-type": "application/json",
            "x-client-request-id": get_client_reqid(),
            "x-request-id": get_reqid(),
        },
        "body": content,
    }



from ark.cli.utils.cr import get_default_cr_host
from ark.cli.utils.tos import TosSDKDownloader
from ark.cli.utils.vefaas import VeFaaS

__all__ = ["TosSDKDownloader", "VeFaaS", "get_default_cr_host"]



import asyncio
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Dict, Generic, Optional, Tuple, Type, Union

import fastapi
import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field
from starlette.responses import StreamingResponse

from ark.core.assistant.middleware import LogIdMiddleware
from ark.core.client import Client, get_client_pool
from ark.core.idl.common_protocol import RequestType, ResponseType
from ark.core.idl.maas_protocol import MaasChatRequest
from ark.core.runtime.asyncio import AsyncRunner
from ark.core.utils.types import load_request


def _get_lock() -> asyncio.Lock:
    return asyncio.Lock()


def _default_endpoint_config() -> Dict[str, Type[MaasChatRequest]]:
    return {"/api/v2/assistant/chat": MaasChatRequest}


def _default_healthcheck_config() -> str:
    return "/healthz"


class AssistantServer(BaseModel, Generic[RequestType, ResponseType]):
    """AssistantServer in charge of the server router and runtime"""

    runner: AsyncRunner
    """ handler runtime of assistant server"""
    clients: Optional[Dict[str, Tuple[Type[Client], Dict[str, Any]]]]
    """
    configuration of client cls & input config
    format: {$client_name: ($cls, $config_dict) }
    """
    endpoint_config: Dict[str, Type[RequestType]] = Field(
        default_factory=_default_endpoint_config
    )
    """
    configuration of endpoint path & request class,
    format: {$ep_path: $request_cls }
    """
    health_check_path: str = Field(default_factory=_default_healthcheck_config)
    """path for server health check"""
    app: FastAPI
    """server application"""

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        runner: AsyncRunner,
        endpoint_config: Optional[Dict[str, Type[RequestType]]] = None,
        clients: Optional[Dict[str, Tuple[Type[Client], Dict[str, Any]]]] = None,
        app: Optional[FastAPI] = None,
        health_check_path: Optional[str] = None,
        **kwargs: Any,
    ):
        @asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncIterator[Dict[str, Any]]:
            yield {"client_pool": get_client_pool(clients)}

        super().__init__(
            runner=runner,
            endpoint_config=endpoint_config or _default_endpoint_config(),
            clients=clients,
            health_check_path=health_check_path or _default_healthcheck_config(),
            app=app or FastAPI(lifespan=lifespan),
            **kwargs,
        )

        self.add_middlewares(self.app)
        self.add_routes(self.app)

    async def handler(self, http_request: fastapi.Request) -> fastapi.Response:
        try:
            request: RequestType = await load_request(
                http_request=http_request,
                req_cls=self.get_request_cls(http_request.url.path),
            )

            if request.stream:
                generator = self.runner.astream(request)
                return StreamingResponse(generator, media_type="text/event-stream")
            else:
                return await self.runner.arun(request)

        except Exception as e:
            raise HTTPException(status_code=400, detail=f"Invalid request: {str(e)}")

    async def health_check(self) -> Any:
        return {}

    def add_routes(self, app: FastAPI) -> None:
        for endpoint_path, request_cls in self.endpoint_config.items():
            app.add_api_route(
                endpoint_path,
                self.handler,
                methods=["POST", "OPTIONS", "GET"],
            )
        app.add_api_route(
            self.health_check_path,
            self.health_check,
            methods=["GET"],
        )

    def get_request_cls(self, api_path: str) -> Type[RequestType]:
        assert api_path in self.endpoint_config, ValueError(
            "no corresponding api in endpoint configs"
        )
        return self.endpoint_config[api_path]

    @staticmethod
    def add_middlewares(app: FastAPI) -> None:
        app.add_middleware(LogIdMiddleware)
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    @staticmethod
    def run(
        app: Union[FastAPI, str],
        host: str = "0.0.0.0",
        port: int = 8080,
        workers_num: int = 1,
    ) -> None:
        uvicorn.run(app, host=host, port=port, workers=workers_num)



import asyncio
from typing import Any, Dict, List, Optional, Union

import aiohttp
import requests

from ark.component.action.function_call import FunctionCall
from ark.component.action.utils import convert_actions, get_ark_client
from ark.core.client import ArkClient
from ark.core.idl.common_protocol import Action, ApiCall, Parameters
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
)
from ark.core.task import task


class ActionCall(FunctionCall):
    actions: Dict[str, Action]
    api_credentials: Optional[Any] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        endpoint_id: str,
        messages: List[MaasChatMessage],
        actions: List[Dict[str, Any]],
        parameters: Optional[Union[Parameters, Dict[str, Any]]] = None,
        client: Optional[ArkClient] = None,
        api_credentials: Optional[Any] = None,
        extra: Optional[Dict[str, str]] = {},
        **kwargs: Any,
    ):
        converted_actions = convert_actions(actions)
        super().__init__(
            endpoint_id=endpoint_id,
            messages=messages,
            parameters=parameters,
            functions=[
                action.tool.function
                for action in converted_actions.values()
                if action.tool.function
            ],
            actions=converted_actions,  # type: ignore
            client=client or get_ark_client(),
            extra=extra,
            **kwargs,
        )
        self.api_credentials = api_credentials

    @task()
    async def async_call_apis(
        self, api_calls: Optional[List[ApiCall]] = None
    ) -> List[Any]:
        if api_calls is None or len(api_calls) == 0:
            return []

        async with aiohttp.ClientSession() as session:
            tasks = []
            for api_call in api_calls:
                if api_call.name not in self.actions:
                    continue

                action = self.actions[api_call.name]
                tasks.append(
                    self._async_call_api(
                        session,
                        action.server.url,
                        action.path,
                        action.http_method,
                        api_call.headers,
                        api_call.parameters,
                        api_call.json_datas,
                    )
                )

            responses = await asyncio.gather(*tasks)
        return responses

    @staticmethod
    async def _async_call_api(
        session: aiohttp.ClientSession,
        base_url: str,
        path: str,
        method: str,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        json_data: Optional[Any] = None,
    ) -> Any:
        url = f"{base_url.rstrip('/')}/{path.lstrip('/')}"
        async with session.request(
            method=method, url=url, headers=headers, params=params, json=json_data
        ) as response:
            return await response.json()

    @task()
    async def call_apis(self, api_calls: Optional[List[ApiCall]] = None) -> List[Any]:
        if api_calls is None or len(api_calls) == 0:
            return []

        responses = []
        with requests.session() as session:
            for api_call in api_calls:
                if api_call.name not in self.actions:
                    continue

                action = self.actions[api_call.name]
                response = self._call_api(
                    session,
                    action.server.url,
                    action.path,
                    action.http_method,
                    api_call.headers,
                    api_call.parameters,
                    api_call.json_datas,
                )
                responses.append(response)
        return responses

    @staticmethod
    def _call_api(
        session: requests.Session,
        base_url: str,
        path: str,
        method: str,
        headers: Optional[Any] = None,
        params: Optional[Any] = None,
        json_data: Optional[Any] = None,
    ) -> Any:
        url = f"{base_url.rstrip('/')}/{path.lstrip('/')}"
        return session.request(
            method, url, headers=headers, params=params, json=json_data
        )



import contextvars
import os
import time
from typing import Any, Dict, Optional, Union

from starlette.datastructures import Headers

from ark.core.idl.common_protocol import ModelDesc

_headers: contextvars.ContextVar[Headers] = contextvars.ContextVar("_headers")


def set_headers(val: Headers) -> None:
    _headers.set(val)


def get_headers(default_val: Headers = Headers()) -> Headers:
    return _headers.get(default_val)


_resource_id: contextvars.ContextVar[str] = contextvars.ContextVar("_resource_id")


def set_resource_id(val: str = "") -> None:
    _resource_id.set(val)


def get_resource_id(default_val: str = "") -> str:
    return _resource_id.get(default_val)


_resource_type: contextvars.ContextVar[str] = contextvars.ContextVar("_resource_type")


def set_resource_type(val: str = "") -> None:
    _resource_type.set(val)


def get_resource_type(default_val: str = "") -> str:
    return _resource_type.get(default_val)


_reqid: contextvars.ContextVar[str] = contextvars.ContextVar("_reqid")


def set_reqid(val: str = "") -> None:
    _reqid.set(val)


def get_reqid(default_val: str = "") -> str:
    return _reqid.get(default_val)


_client_reqid: contextvars.ContextVar = contextvars.ContextVar("_client_reqid")


def set_client_reqid(val: str) -> None:
    _client_reqid.set(val)


def get_client_reqid(default_val: str = "") -> str:
    return _client_reqid.get(default_val)


_model_info: contextvars.ContextVar[ModelDesc] = contextvars.ContextVar("_model_info")


def set_model_info(model_info: ModelDesc) -> None:
    _model_info.set(model_info)


def get_model_info() -> Union[ModelDesc, None]:
    return _model_info.get(None)


_start_time: contextvars.ContextVar[float] = contextvars.ContextVar("_start_time")


def set_start_time(start_time: float) -> None:
    _start_time.set(start_time)


def get_start_time() -> float:
    return _start_time.get(time.time())


_account_id: contextvars.ContextVar[str] = contextvars.ContextVar("_account_id")


def set_account_id(val: str = "") -> None:
    _account_id.set(val)


def get_account_id(default_val: str = "") -> str:
    return _account_id.get(default_val)


_custom_attributes: contextvars.ContextVar[
    Optional[Dict[str, Any]]
] = contextvars.ContextVar("_custom_attributes")


def set_custom_attributes(val: Dict[str, Any]) -> None:
    _custom_attributes.set(val)


def get_custom_attributes() -> Optional[Dict[str, Any]]:
    return _custom_attributes.get(None)


_req_source_type: contextvars.ContextVar[str] = contextvars.ContextVar(
    "_req_source_type"
)


def set_req_source_type(val: str = "") -> None:
    _req_source_type.set(val)


def get_req_source_type(default_val: str = "") -> str:
    return _req_source_type.get(default_val)


HEADERS_WHITE_LIST = {
    "authorization": "Authorization",
    "x-account-id": "X-Account-Id",
    "x-user-id": "X-User-Id",
    "x-project-name": "X-Project-Name",
}


def get_extra_headers(extra_headers: Optional[Dict[str, str]] = {}) -> Dict[str, str]:
    if not extra_headers:
        extra_headers = {}
    if os.getenv("EXPOSE_HEADERS") == "true":
        headers = get_headers()
        for header in headers.keys():
            if header.lower() in HEADERS_WHITE_LIST:
                extra_headers[HEADERS_WHITE_LIST[header.lower()]] = headers[header]

    extra_headers["X-Tt-Logid"] = get_reqid()
    extra_headers["X-Request-Id"] = get_reqid()
    extra_headers["X-Client-Request-Id"] = get_client_reqid()
    extra_headers["Ark-Origin-Service-Type"] = "bot-procode"
    extra_headers["Ark-Origin-Service-Id"] = get_resource_id()
    # for procode difference
    extra_headers["Ark-Origin-Request-Id"] = get_reqid()
    return extra_headers



from ark.component.v3.llm.rag_intention.rag_intention import (
    AbstractIntention,
    KnowledgeIntention,
    RagIntention,
)

__all__ = ["RagIntention", "KnowledgeIntention", "AbstractIntention"]
